# Module containing Linear Solver classes
from math import isnan
import numpy as np

from occamypy.solver import Solver
from occamypy import problem as P

zero = 10 ** (np.floor(np.log10(np.abs(float(np.finfo(np.float64).tiny)))) + 2)  # Check for avoid Overflow or Underflow


class CG(Solver):
    """Linear-Conjugate Gradient and Steepest-Descent Solver parent object"""

    # Default class methods/functions
    def __init__(self, stopper, steepest=False, logger=None):
        """
        Constructor for LCG/SD Solver:
        :param stopper: Stopper, object to terminate inversion
        :param steepest: bool, use the steepest-descent instead of conjugate gradient [False]
        :param logger: Logger, object to write inversion log file [None]
        """
        # Calling parent construction
        super(CG, self).__init__()
        # Defining stopper object
        self.stopper = stopper
        # Whether to run steepest descent or not
        self.steepest = steepest
        # Logger object to write on log file
        self.logger = logger
        # Overwriting logger of the Stopper object
        self.stopper.logger = self.logger
        # print formatting
        self.iter_msg = "iter = %s, obj = %.5e, resnorm = %.2e, gradnorm = %.2e, feval = %d"

    def __del__(self):
        """Default destructor"""
        return

    def run(self, problem, verbose=False, restart=False):
        """Running LCG and steepest-descent solver"""
        self.create_msg = verbose or self.logger

        # Resetting stopper before running the inversion
        self.stopper.reset()
        # Check for preconditioning
        precond = True if "prec" in dir(problem) and problem.prec is not None else False

        if not restart:
            if self.create_msg:
                msg = 90 * "#" + "\n"
                msg += "\t\t\t\tPRECONDITIONED " if precond else "\t\t\t\t"
                msg += "LINEAR %s SOLVER\n" % ("STEEPEST-DESCENT log file" if self.steepest else "CONJUGATE GRADIENT log file")
                msg += "\tRestart folder: %s\n" % self.restart.restart_folder
                msg += "\tModeling Operator:\t\t%s\n" % problem.op
                msg += 90 * "#" + "\n"
                if verbose:
                    print(msg.replace(" log file", ""))
                if self.logger:
                    self.logger.addToLog(msg)

            # Setting internal vectors (model and search direction vectors)
            prblm_mdl = problem.get_model()
            cg_mdl = prblm_mdl.clone()
            cg_dmodl = prblm_mdl.clone().zero()

            # Other internal variables
            iiter = 0
        else:
            # Retrieving parameters and vectors to restart the solver
            if self.create_msg:
                msg = "Restarting previous solve run from: %s" % self.restart.restart_folder
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
            self.restart.read_restart()
            iiter = self.restart.retrieve_parameter("iter")
            initial_obj_value = self.restart.retrieve_parameter("obj_initial")
            cg_mdl = self.restart.retrieve_vector("cg_mdl")
            cg_dmodl = self.restart.retrieve_vector("cg_dmodl")
            if not precond:
                cg_dres = self.restart.retrieve_vector("cg_dres")
            else:
                dot_grad_prec_grad = self.restart.retrieve_vector("dot_grad_prec_grad")
            # Setting the model and residuals to avoid residual twice computation
            problem.set_model(cg_mdl)
            prblm_mdl = problem.get_model()
            # Setting residual vector to avoid its unnecessary computation
            problem.set_residual(self.restart.retrieve_vector("prblm_res"))

        # Common variables unrelated to restart
        success = True
        # Variables necessary to return inverted model if inversion stops earlier
        prev_mdl = prblm_mdl.clone().zero()
        if precond:
            cg_prec_grad = cg_dmodl.clone().zero()

        # Iteration loop
        while True:
            # Computing objective function
            obj0 = problem.get_obj(cg_mdl)  # Compute objective function value
            prblm_res = problem.get_res(cg_mdl)  # Compute residuals
            prblm_grad = problem.get_grad(cg_mdl)  # Compute the gradient
            if iiter == 0:
                initial_obj_value = obj0  # For relative objective function value
                # Saving initial objective function value
                self.restart.save_parameter("obj_initial", initial_obj_value)
                if self.create_msg:
                    msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                           obj0,
                                           problem.get_rnorm(cg_mdl),
                                           problem.get_gnorm(cg_mdl),
                                           problem.get_fevals())
                    if verbose:
                        print(msg)
                    # Writing on log file
                    if self.logger:
                        self.logger.addToLog(msg)
                # Check if either objective function value or gradient norm is NaN
                if isnan(obj0) or isnan(prblm_grad.norm()):
                    raise ValueError("Either gradient norm or objective function value NaN!")
                # Set internal delta residual vector
                if not precond:
                    cg_dres = prblm_res.clone().zero()
            if prblm_grad.norm() == 0.:
                print("Gradient vanishes identically")
                break

            # Saving results
            self.save_results(iiter, problem, force_save=False)
            prev_mdl.copy(prblm_mdl)  # Keeping the previous model

            # Computing alpha and beta coefficients
            if precond:
                # Applying preconditioning to current gradient
                problem.prec.forward(False, prblm_grad, cg_prec_grad)
                if iiter == 0 or self.steepest:
                    # Steepest descent
                    beta = 0.
                    dot_grad_prec_grad = prblm_grad.dot(cg_prec_grad)
                else:
                    # Conjugate-gradient coefficients for preconditioned CG
                    dot_grad_prec_grad_old = dot_grad_prec_grad
                    if dot_grad_prec_grad_old == 0.:
                        success = False
                        # Writing on log file
                        if self.logger:
                            self.logger.addToLog("Gradient orthogonal to preconditioned one, will terminate solver")
                    dot_grad_prec_grad = prblm_grad.dot(cg_prec_grad)
                    beta = dot_grad_prec_grad / dot_grad_prec_grad_old
                # Update search direction
                cg_dmodl.scaleAdd(cg_prec_grad, beta, 1.0)
                cg_dmodld = problem.get_dres(cg_mdl, cg_dmodl)  # Project search direction into the data space
                dot_cg_dmodld = cg_dmodld.dot(cg_dmodld)
                if dot_cg_dmodld == 0.0:
                    success = False
                    # Writing on log file
                    if self.logger:
                        self.logger.addToLog(
                            "Search direction orthogonal to span of linear operator, will terminate solver")
                else:
                    alpha = - dot_grad_prec_grad / dot_cg_dmodld
                    # Writing on log file
                    if beta == 0.:
                        msg = "Steepest-descent step length: %.2e" % alpha
                    else:
                        msg = "Conjugate alpha, beta: %.2e, %.2e" % (alpha, beta)
                    if self.logger:
                        self.logger.addToLog(msg)
            else:
                prblm_gradd = problem.get_dres(cg_mdl, prblm_grad)  # Project gradient into the data space
                # Computing alpha and beta coefficients
                if iiter == 0 or self.steepest:
                    # Steepest descent
                    beta = 0.0
                    dot_gradd = prblm_gradd.dot(prblm_gradd)
                    if dot_gradd <= zero:
                        success = False
                        # Writing on log file
                        if self.logger:
                            self.logger.addToLog(
                                "Gradient orthogonal to span of linear operator, will terminate solver")
                    else:
                        dot_gradd_res = prblm_gradd.dot(prblm_res)
                        alpha = - np.real(dot_gradd_res) / dot_gradd
                        msg = "Steppest-descent step length: " + str(alpha)
                        # Writing on log file
                        if iiter == 0:
                            msg = "First steppest-descent step length: " + str(alpha)
                        if self.logger:
                            self.logger.addToLog(msg)
                else:
                    # Conjugate-gradient coefficients
                    dot_gradd = prblm_gradd.dot(prblm_gradd)
                    dot_dres = cg_dres.dot(cg_dres)
                    dot_gradd_dres = np.real(prblm_gradd.dot(cg_dres))
                    if dot_gradd <= zero or dot_dres <= zero:
                        success = False
                    else:
                        determ = dot_gradd * dot_dres - dot_gradd_dres * dot_gradd_dres
                        # Checking if alpha or beta are becoming infinity
                        if abs(determ) < zero:
                            if self.create_msg:
                                msg = "Plane-search method fails (zero det: %.2e), will terminate solver" % determ
                                if verbose:
                                    print(msg)
                                if self.logger:
                                    self.logger.addToLog(msg)
                            break
                        dot_gradd_res = np.real(prblm_gradd.dot(prblm_res))
                        dot_dres_res = np.real(cg_dres.dot(prblm_res))
                        alpha = -(dot_dres * dot_gradd_res - dot_gradd_dres * dot_dres_res) / determ
                        beta = (dot_gradd_dres * dot_gradd_res - dot_gradd * dot_dres_res) / determ
                        # Writing on log file
                        if self.logger:
                            self.logger.addToLog("Conjugate alpha,beta: " + str(alpha) + ", " + str(beta))

            if not success:
                if self.create_msg:
                    msg = "Stepper couldn't find a proper step size, will terminate solver"
                    if verbose:
                        print(msg)
                    if self.logger:
                        self.logger.addToLog(msg)
                break

            if precond:
                # modl = modl + alpha * dmodl
                cg_mdl.scaleAdd(cg_dmodl, 1.0, alpha)  # Update model
            else:
                # dmodl = alpha * grad + beta * dmodl
                cg_dmodl.scaleAdd(prblm_grad, beta, alpha)  # update search direction
                # modl = modl + dmodl
                cg_mdl.scaleAdd(cg_dmodl)  # Update model

            # Increasing iteration counter
            iiter += 1
            # Setting the model
            problem.set_model(cg_mdl)
            # Projecting model onto the bounds (if any)
            if "bounds" in dir(problem):
                problem.bounds.apply(cg_mdl)

            if prblm_mdl.isDifferent(cg_mdl):
                # Model went out of the bounds
                msg = "Model hit provided bounds. Projecting it onto them."
                if self.logger:
                    self.logger.addToLog(msg)
                # Recomputing m_current = m_new - dmodl
                prblm_mdl.scaleAdd(cg_dmodl, 1.0, -1.0)
                # Finding the projected dmodl = m_new_clipped - m_current
                cg_dmodl.copy(cg_mdl)
                cg_dmodl.scaleAdd(prblm_mdl, 1.0, -1.0)
                problem.set_model(cg_mdl)
                if precond:
                    cg_dmodl.scale(1.0 / alpha)  # Unscaling the search direction
                else:
                    # copying previous residuals dres = res_old
                    cg_dres.copy(prblm_res)
                    # Computing actual change in the residual vector dres = res_new - res_old
                    prblm_res = problem.get_res(cg_mdl)  # New residual vector
                    cg_dres.scaleAdd(prblm_res, -1.0, 1.0)
            else:
                # Setting residual vector to avoid its unnecessary computation (if model was not clipped)
                if precond:
                    # res = res + alpha * dres
                    prblm_res.scaleAdd(cg_dmodld, 1.0, alpha)  # Update residuals
                else:
                    # dres  = alpha * gradd + beta * dres
                    cg_dres.scaleAdd(prblm_gradd, beta, alpha)  # Update residual step
                    # res = res + dres
                    prblm_res.scaleAdd(cg_dres)  # Update residuals
                problem.set_residual(prblm_res)

            # Computing new objective function value
            obj1 = problem.get_obj(cg_mdl)
            if obj1 >= obj0:
                if self.create_msg:
                    msg = "Objective function didn't reduce, will terminate solver:\n\t" \
                          "obj_new = %.5e\tobj_cur = %.5e" % (obj1, obj0)
                    if verbose:
                        print(msg)
                    # Writing on log file
                    if self.logger:
                        self.logger.addToLog(msg)
                problem.set_model(prev_mdl)
                break

            # Saving current model and previous search direction in case of restart
            self.restart.save_parameter("iter", iiter)
            self.restart.save_vector("cg_mdl", cg_mdl)
            self.restart.save_vector("cg_dmodl", cg_dmodl)
            # Saving data space vectors or scaling if preconditioned
            if not precond:
                self.restart.save_vector("cg_dres", cg_dres)
            else:
                self.restart.save_parameter("dot_grad_prec_grad", dot_grad_prec_grad)
            self.restart.save_vector("prblm_res", prblm_res)

            # iteration info
            if self.create_msg:
                msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                       obj1,
                                       problem.get_rnorm(cg_mdl),
                                       problem.get_gnorm(cg_mdl),
                                       problem.get_fevals())
                if verbose:
                    print(msg)
                # Writing on log file
                if self.logger:
                    self.logger.addToLog("\n" + msg)
            # Check if either objective function value or gradient norm is NaN
            if isnan(obj1) or isnan(prblm_grad.norm()):
                raise ValueError("Either gradient norm or objective function value NaN!")
            if self.stopper.run(problem, iiter, initial_obj_value, verbose):
                break

        # Writing last inverted model
        self.save_results(iiter, problem, force_save=True, force_write=True)
        if self.create_msg:
            msg = 90 * "#" + "\n"
            msg += "\t\t\t\tPRECONDITIONED " if precond else "\t\t\t\t"
            msg += "LINEAR %s SOLVER log file end\n" % ("STEEPEST-DESCENT" if self.steepest else "CONJUGATE GRADIENT")
            msg += 90 * "#" + "\n"
            if verbose:
                print(msg.replace(" log file", ""))
            if self.logger:
                self.logger.addToLog(msg)
        # Clear restart object
        self.restart.clear_restart()

        return


class SD(CG):
    def __init__(self, stopper, logger=None):
        super(SD, self).__init__(stopper, steepest=True, logger=logger)
    
    
def _sym_ortho(a, b):
    """
    Stable implementation of Givens rotation.
    Notes
    -----
    The routine 'SymOrtho' was added for numerical stability. This is
    recommended by S.-C. Choi in [1]_.  It removes the unpleasant potential of
    ``1/eps`` in some important places (see, for example text following
    "Compute the next plane rotation Qk" in minres.py).
    References
    ----------
    .. [1] S.-C. Choi, "Iterative Methods for Singular Linear Equations
           and Least-Squares Problems", Dissertation,
           http://www.stanford.edu/group/SOL/dissertations/sou-cheng-choi-thesis.pdf
    """
    if b == 0.:
        return np.sign(a), 0, np.abs(a)
    elif a == 0.:
        return 0., np.sign(b), np.abs(b)
    elif np.abs(b) > np.abs(a):
        tau = a / b
        s = np.sign(b) / np.sqrt(1. + tau * tau)
        c = s * tau
        r = b / s
    else:
        tau = b / a
        c = np.sign(a) / np.sqrt(1. + tau * tau)
        s = c * tau
        r = a / c
    return c, s, r


class LSQR(Solver):
    """
       LSQR Solver parent object following algorithm in Paige and Saunders (1982)
       Find the least-squares solution to a large, sparse, linear system
       of equations.
       The function solves Ax = b or min ||b - Ax||^2
       If A is symmetric, LSQR should not be used! Use SymLCGsolver
       If initial model different than zero the solver will perform the following:
         1. Compute initial residual vector ``r0 = b - A*x0``.
         2. Use LSQR to solve the system  ``A*dx = r0``.
         3. Add the correction dx to obtain a final solution ``x = x0 + dx``.
    """

    def __init__(self, stopper, estimate_cond=False, estimate_var=False, logger=None):
        """
        Constructor for LSQR Solver:
        :param stopper          : Stopper, object to terminate inversion
        :param estimate_cond    : Boolean, whether the condition number of A is estimated
                                  access self.acond after solver run [False]
        :param estimate_var     : Boolean, whether the diagonal of A'A^-1 is estimated or not;
                                  access self.var after solver run [False]
        :param logger           : Logger, object to write inversion log file [None]
        """
        # Calling parent construction
        super(LSQR, self).__init__()
        # Defining stopper object
        self.stopper = stopper
        # Logger object to write on log file
        self.logger = logger
        # Create var variable if estimate var is requested
        self.est_cond = True if estimate_cond or estimate_var else False
        self.var = estimate_var
        # Overwriting logger of the Stopper object
        self.stopper.logger = self.logger
        # print formatting
        self.iter_msg = "iter = %s, obj = %.5e, resnorm = %.2e, gradnorm = %.2e, feval = %d"

    def __del__(self):
        """Default destructor"""
        return

    def run(self, problem, verbose=False, restart=False):
        """Running LSQR solver"""
        self.create_msg = verbose or self.logger

        # Resetting stopper before running the inversion
        self.stopper.reset()

        # Setting internal vectors and initial variables
        prblm_mdl = problem.get_model()
        initial_mdl = prblm_mdl.clone()

        if not restart:
            if self.create_msg:
                msg = 90 * "#" + "\n"
                msg += "\t\t\t\tLSQR SOLVER log file\n"
                msg += "\tRestart folder: %s\n" % self.restart.restart_folder
                msg += "\tModeling Operator:\t\t%s\n" % problem.op
                msg += 90 * "#" + "\n"
                if verbose:
                    print(msg.replace("log file", ""))
                if self.logger:
                    self.logger.addToLog(msg)

            # If initial model different than zero the solver will perform the following:
            # 1. Compute a residual vector ``r0 = b - A*x0``.
            # 2. Use LSQR to solve the system  ``A*dx = r0``.
            # 3. Add the correction dx to obtain a final solution ``x = x0 + dx``.
            prblm_res = problem.get_res(initial_mdl)  # Initial data residuals
            obj0 = initial_obj_value = problem.get_obj(initial_mdl)  # For relative objective function value
            # Saving initial objective function value
            u = prblm_res.clone().scale(-1.0)
            x = prblm_mdl.clone().zero()  # Solution vector
            w = x.clone()
            v = x.clone()
            if self.est_cond:
                dk = x.clone()
                ddnorm = 0.
            # Estimating variance or diagonal elements of the inverse
            if self.var:
                self.var = x.clone()

            # Other internal variables
            iiter = 0

            # Initial inversion parameters
            alpha = 0.
            beta = u.norm()

            if beta > 0.:
                u.scale(1. / beta)
                # A.H * u => gradient with scaled residual vector
                problem.set_model(x)  # x = 0
                problem.set_residual(u)  # res = u
                prblm_grad = problem.get_grad(x)  # g = A.H * u
                v.copy(prblm_grad)  # v = g
                alpha = v.norm()
            else:
                prblm_grad = problem.get_grad(initial_mdl)
            if alpha > 0.:
                v.scale(1. / alpha)
                w.copy(v)

            rhobar = alpha
            phibar = beta
            anorm = 0.

            # First inversion logging
            self.restart.save_parameter("obj_initial", initial_obj_value)
            # Estimating residual and gradient norms
            prblm_res.scale(phibar)
            prblm_grad.scale(alpha*beta)
            if self.create_msg:
                msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                       initial_obj_value,
                                       problem.get_rnorm(prblm_mdl),
                                       problem.get_gnorm(prblm_mdl),
                                       problem.get_fevals())
                if verbose:
                    print(msg)
                # Writing on log file
                if self.logger:
                    self.logger.addToLog(msg)
            # Check if either objective function value or gradient norm is NaN
            if isnan(initial_obj_value) or isnan(problem.get_gnorm(x)):
                raise ValueError("Either gradient norm or objective function value NaN!")

        else:
            # Retrieving parameters and vectors to restart the solver
            if self.create_msg:
                msg = "Restarting previous solve run from: %s" % self.restart.restart_folder
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
            self.restart.read_restart()

            # Retrieving iteration number
            iiter = self.restart.retrieve_parameter("iter")
            initial_obj_value = self.restart.retrieve_parameter("obj_initial")
            obj0 = self.restart.retrieve_parameter("obj0")
            # Retrieving state vectors
            u = self.restart.retrieve_vector("u")
            x = self.restart.retrieve_vector("x")
            w = self.restart.retrieve_vector("w")
            v = self.restart.retrieve_vector("v")
            problem.set_model(x)
            problem.set_residual(u)

            prblm_res = problem.get_res(x)
            if self.est_cond:
                dk = self.restart.retrieve_vector("dk")
                ddnorm = self.restart.retrieve_parameter("ddnorm")
            if self.var:
                self.var = self.restart.retrieve_vector("var")
            # Retrieving inversion parameters
            alpha = self.restart.retrieve_parameter("alpha")
            rhobar = self.restart.retrieve_parameter("rhobar")
            phibar = self.restart.retrieve_parameter("phibar")
            anorm = self.restart.retrieve_parameter("anorm")

        # Common variables unrelated to restart
        prblm_mdl = problem.get_model()
        inv_model = prblm_mdl.clone()  # Inverted model to be saved during the inversion
        # Variables necessary to return inverted model if inversion stops earlier
        prev_x = x.clone().zero()
        early_stop =False

        # Iteration loop
        while True:
            if problem.get_gnorm(prblm_mdl) == 0.:
                print("Gradient vanishes identically")
                break

            # Saving results
            inv_model.copy(initial_mdl)
            inv_model.scaleAdd(x)  # x = x0 + dx; Updating inverted model
            self.save_results(iiter, problem, model=inv_model, force_save=False)
            # Necessary to save previous iteration
            prev_x.copy(x)

            """
                %     Perform the next step of the bidiagonalization to obtain the
                %     next  beta, u, alpha, v.  These satisfy the relations
                %                beta*u  =  A*v   -  alpha*u,
                %                alpha*v  =  A'*u  -  beta*v.
            """


            # A.matvec(v) (i.e., projection of v onto the data space)
            v_prblm = problem.get_dres(x, v)
            # u = A.matvec(v) - alpha * u
            u.scaleAdd(v_prblm, -alpha, 1.0)
            beta = u.norm()

            if beta > 0.:
                u.scale(1. / beta)
                anorm = np.sqrt(anorm ** 2 + alpha ** 2 + beta ** 2)
                problem.set_model(x)
                problem.set_residual(u)  # res = u
                prblm_grad = problem.get_grad(x)  # g = A.H * u
                # v = A.rmatvec(u) - beta * v
                v.scaleAdd(prblm_grad, -beta, 1.)
                alpha = v.norm()
                if alpha > 0.:
                    v.scale(1. / alpha)
            else:
                problem.set_model(x)
                problem.set_residual(u)  # res = u

            # Use a plane rotation to eliminate the subdiagonal element (beta)
            # of the lower-bidiagonal matrix, giving an upper-bidiagonal matrix.
            cs, sn, rho = _sym_ortho(rhobar, beta)

            theta = sn * alpha
            rhobar = -cs * alpha
            phi = cs * phibar
            phibar *= sn

            # Estimating residual and gradient norms
            prblm_res.scale(phibar)
            if prblm_grad.norm() > zero:
                prblm_grad.scale(alpha * abs(sn * phi) / prblm_grad.norm())
            else:
                prblm_grad.zero()
            # New objective function value
            obj1 = problem.get_obj(prblm_mdl)


            # Update x and w.
            # x = x + t1 * w
            x.scaleAdd(w, 1.0, phi / rho)
            # w = v + t2 * w
            w.scaleAdd(v, -theta / rho, 1.0)

            # Increasing iteration counter
            iiter += 1

            # Checking new objective function value
            if obj1 >= obj0:
                if self.create_msg:
                    msg = "Objective function didn't reduce, will terminate solver:\n\t" \
                          "obj_new = %.5e\tobj_cur = %.5e" % (obj1, obj0)
                    if verbose:
                        print(msg)
                    # Writing on log file
                    if self.logger:
                        self.logger.addToLog(msg)
                early_stop = True
                break

            if self.est_cond:
                dk.copy(w).scale(1. / rho)
                ddnorm += dk.norm() ** 2
                # Estimate the condition of the matrix  Abar,
                self.acond = anorm * np.sqrt(ddnorm)
                self.restart.save_vector("dk", dk)
                self.restart.save_parameter("ddnorm", ddnorm)
            if self.var:
                # var = var + dk ** 2
                self.var.scaleAdd(dk.clone().multiply(dk))
                self.restart.save_vector("var", self.var)

            # Saving previous objective function value
            obj0 = obj1

            # Saving state variables and vectors for restart
            self.restart.save_parameter("iter", iiter)
            self.restart.save_parameter("obj0", obj0)
            self.restart.save_parameter("alpha", alpha)
            self.restart.save_parameter("rhobar", rhobar)
            self.restart.save_parameter("phibar", phibar)
            self.restart.save_parameter("anorm", anorm)
            self.restart.save_vector("u", u)
            self.restart.save_vector("x", x)
            self.restart.save_vector("w", w)
            self.restart.save_vector("v", v)

            # iteration info
            if self.create_msg:
                msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                       problem.get_obj(prblm_mdl),
                                       problem.get_rnorm(prblm_mdl),
                                       problem.get_gnorm(prblm_mdl),
                                       problem.get_fevals())
                if self.est_cond:
                    msg += ", condition_num =  %.2e, matrix_norm = %.2e" % (self.acond, anorm)
                if verbose:
                    print(msg)
                # Writing on log file
                if self.logger:
                    self.logger.addToLog("\n" + msg)
            # Check if either objective function value or gradient norm is NaN
            if isnan(problem.get_obj(prblm_mdl)) or isnan(problem.get_gnorm(prblm_mdl)):
                raise ValueError("Either gradient norm or objective function value NaN!")
            if self.stopper.run(problem, iiter, initial_obj_value, verbose):
                break

        # Writing last inverted model
        inv_model.copy(initial_mdl)
        if early_stop:
            x.copy(prev_x)
        inv_model.scaleAdd(x)  # x = x0 + dx; Updating inverted model
        self.save_results(iiter, problem, model=inv_model, force_save=True, force_write=True)
        prblm_mdl.copy(inv_model) # Setting inverted model to final one
        if self.create_msg:
            msg = 90 * "#" + "\n"
            msg += "\t\t\t\tLSQR SOLVER log file end\n"
            msg += 90 * "#" + "\n"
            if verbose:
                print(msg.replace("log file ", ""))
            if self.logger:
                self.logger.addToLog(msg)
        # Clear restart object
        self.restart.clear_restart()
        return


class CGsym(Solver):
    """Linear-Conjugate Gradient Solver (for symmetric systems) parent object"""

    # Default class methods/functions
    def __init__(self, stopper, steepest=False, logger=None):
        """Constructor for LCG Solver for symmetric systems"""
        # Calling parent construction
        super(CGsym, self).__init__()
        # Defining stopper object
        self.stopper = stopper
        # Whether to run steepest descent or not
        self.steepest = steepest
        # Logger object to write on log file
        self.logger = logger
        # Overwriting logger of the Stopper object
        self.stopper.logger = self.logger
        # Setting defaults for saving results
        self.setDefaults()
        # print formatting
        self.iter_msg = "iter = %s, obj = %.5e, resnorm = %.2e, feval = %d"
        return

    def __del__(self):
        """Default destructor"""
        return

    def run(self, problem, verbose=False, restart=False):
        """Running LCG solver for symmetric systems"""
        self.create_msg = verbose or self.logger

        # Resetting stopper before running the inversion
        self.stopper.reset()
        # Checking if we are solving a linear square problem
        if not isinstance(problem, P.LeastSquaresSymmetric):
            raise TypeError("ERROR! Provided problem object not a linear symmetric problem")
        # Check for preconditioning
        precond = False
        if "prec" in dir(problem):
            if problem.prec is not None:
                precond = True
        if not restart:
            if self.create_msg:
                msg = 90 * "#" + "\n"
                msg += "PRECONDITIONED " if precond else ""
                msg += "LINEAR %s SOLVER FOR SYMMETRIC MATRIX log file" % (
                    "STEEPEST-DESCENT" if self.steepest else "CONJUGATE GRADIENT")
                msg += "Restart folder: %s\n" % self.restart.restart_folder
                msg = 90 * "#" + "\n"
                if verbose:
                    print(msg.replace("log file", ""))
                if self.logger:
                    self.logger.addToLog(msg)

            # Setting internal vectors (model and search direction vectors)
            prblm_mdl = problem.get_model()
            cg_mdl = prblm_mdl.clone()
            cg_dmodl = prblm_mdl.clone().zero()
            if precond:
                cg_prec_res = cg_dmodl.clone()

            # Other internal variables
            iiter = 0
            beta = 0.
        else:
            # Retrieving parameters and vectors to restart the solver
            if self.create_msg:
                msg = "Restarting previous solver run from: %s" % self.restart.restart_folder
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
            self.restart.read_restart()
            iiter = self.restart.retrieve_parameter("iter")
            beta = self.restart.retrieve_parameter("beta")
            obj_old = self.restart.retrieve_parameter("obj_old")
            cg_mdl = self.restart.retrieve_vector("cg_mdl")
            cg_dmodl = self.restart.retrieve_vector("cg_dmodl")
            if precond: cg_prec_res = self.restart.retrieve_vector("cg_prec_res")
            # Setting the model and residuals to avoid residual double computation
            problem.set_model(cg_mdl)
            prblm_mdl = problem.get_model()
            # Setting residual vector to avoid its unnecessary computation
            problem.set_residual(self.restart.retrieve_vector("prblm_res"))

        # Common variables unrelated to restart
        success = True
        data_norm = problem.data.norm()
        prev_mdl = prblm_mdl.clone().zero()

        # Iteration loop
        while True:
            # Computing objective function
            prblm_res = problem.get_res(cg_mdl)  # Compute residuals
            obj0 = problem.get_obj(cg_mdl)  # Compute objective function value
            if iiter == 0:
                if self.create_msg:
                    msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                           obj0,
                                           problem.get_rnorm(cg_mdl),
                                           problem.get_fevals())
                    if verbose:
                        print(msg)
                    msg += "\nrelative data matching (i.e., 1-|Am-b|/|b|): %.1f" \
                           % ((1.0 - problem.get_rnorm(cg_mdl) / data_norm) * 100.0) + "%"
                    # Writing on log file
                    if self.logger:
                        self.logger.addToLog(msg)
                # Check if either objective function value or gradient norm is NaN
                if isnan(obj0):
                    raise ValueError("Error! Objective function value NaN!")

            # Saving results
            self.save_results(iiter, problem, force_save=False)
            # Copying current model in case of early stop
            prev_mdl.copy(cg_mdl)

            # Applying preconditioning to gradient (first time)
            if iiter == 0 and precond:
                problem.prec.forward(False, prblm_res, cg_prec_res)
            # dmodl = beta * dmodl - res
            cg_dmodl.scaleAdd(cg_prec_res if precond else prblm_res, beta, -1.0)  # Update search direction
            prblm_ddmodl = problem.get_dres(cg_mdl, cg_dmodl)  # Project search direction in the data space

            dot_dmodl_ddmodl = cg_dmodl.dot(prblm_ddmodl)
            if precond:
                dot_res = prblm_res.dot(cg_prec_res)  # Dot product of residual and preconditioned one
            else:
                dot_res = prblm_res.dot(prblm_res)
            if dot_res == 0.:
                if self.create_msg:
                    msg = "Residual/Gradient vector vanishes identically"
                    if verbose:
                        print(msg)
                    # Writing on log file
                    if self.logger:
                        self.logger.addToLog(msg)
                break
            elif dot_dmodl_ddmodl == 0.:
                success = False
                msg = "Residual/Gradient vector orthogonal to span of linear operator, will terminate solver"
                # Writing on log file
                if self.logger:
                    self.logger.addToLog(msg)

            if not success:
                if self.create_msg:
                    msg = "Stepper couldn't find a proper step size, will terminate solver"
                    if verbose:
                        print(msg)
                    if self.logger:
                        self.logger.addToLog(msg)
                break

            alpha = dot_res / dot_dmodl_ddmodl
            if self.logger:
                self.logger.addToLog("Alpha step length: %.2e" % alpha)

            # modl = modl + alpha * dmodl
            cg_mdl.scaleAdd(cg_dmodl, sc2=alpha)  # Update model

            # Increasing iteration counter
            iiter = iiter + 1
            # Setting the model and residuals to avoid residual twice computation
            problem.set_model(cg_mdl)

            # Projecting model onto the bounds (if any)
            if "bounds" in dir(problem):
                problem.bounds.apply(cg_mdl)

            if prblm_mdl.isDifferent(cg_mdl):
                # Model went out of the bounds
                msg = "Model hit provided bounds. Projecting it onto them."
                if self.logger:
                    self.logger.addToLog(msg)
                # Recomputing m_current = m_new - alpha * dmodl
                prblm_mdl.scaleAdd(cg_dmodl, 1.0, -alpha)
                # Finding the projected dmodl = m_new_clipped - m_current
                cg_dmodl.copy(cg_mdl)
                cg_dmodl.scaleAdd(prblm_mdl, 1.0, -1.0)
                # dmodl is scaled by the inverse of the step length
                cg_dmodl.scale(1.0 / alpha)
                # copying previos residuals dres = res_old
                prblm_ddmodl.copy(prblm_res)
                problem.set_model(cg_mdl)
                # Computing actual change in the residual vector dres = res_new - res_old
                prblm_res = problem.get_res(cg_mdl)  # New residual vector
                prblm_ddmodl.scaleAdd(prblm_res, -1.0, 1.0)
                # dres is scaled by the inverse of the step length
                prblm_ddmodl.scale(1.0 / alpha)
            else:
                # Setting residual vector to avoid its unnecessary computation (if model was not clipped)
                # res  = res + alpha * dres =  res + alpha * A * dmodl
                prblm_res.scaleAdd(prblm_ddmodl, sc2=alpha)  # update residuals
                problem.set_residual(prblm_res)
                if iiter == 1:
                    problem.fevals += 1  # To correct objective function evaluation number since residuals are set

            # Computing new objective function value
            obj1 = problem.get_obj(cg_mdl)
            if precond:
                # Applying preconditioning to gradient
                problem.prec.forward(False, prblm_res, cg_prec_res)
                dot_res_new = prblm_res.dot(cg_prec_res)
            else:
                # New residual norm
                dot_res_new = prblm_res.dot(prblm_res)
            if not self.steepest:
                beta = dot_res_new / dot_res
            # Checking monotonic behavior of objective function
            if iiter == 1:
                obj_old = obj0  # Saving objective function at iter-1
            else:
                # If not monotonically changing stop the inversion
                if not ((obj_old < obj0 < obj1) or (obj_old > obj0 > obj1)):
                    if self.create_msg:
                        msg = "Objective function variation not monotonic, will terminate solver:" \
                              "obj_old=%.5e obj_cur=%.5e obj_new=%.5e" % (obj_old, obj0, obj1)
                        if verbose:
                            print(msg)
                        # Writing on log file
                        if self.logger:
                            self.logger.addToLog(msg)
                    problem.set_model(prev_mdl)
                    break
                obj_old = obj0  # Saving objective function at iter-1

            # Saving current model and previous search direction in case of restart
            self.restart.save_parameter("iter", iiter)
            self.restart.save_parameter("beta", beta)
            self.restart.save_parameter("obj_old", obj_old)
            self.restart.save_vector("cg_mdl", cg_mdl)
            self.restart.save_vector("cg_dmodl", cg_dmodl)
            if precond:
                self.restart.save_vector("cg_prec_res", cg_prec_res)
            # Saving data space vectors
            self.restart.save_vector("prblm_res", prblm_res)

            # iteration info
            if self.create_msg:
                msg = self.iter_msg % (str(iiter).zfill(self.stopper.zfill),
                                       obj1,
                                       problem.get_rnorm(cg_mdl),
                                       problem.get_fevals())
                if verbose:
                    print(msg)
                msg += "\nrelative data matching (i.e., 1-|Am-b|/|b|): %.1f" \
                       % ((1.0 - problem.get_rnorm(cg_mdl) / data_norm) * 100.0) + "%"
                # Writing on log file
                if self.logger:
                    self.logger.addToLog(msg)
            # Check if either objective function value or gradient norm is NaN
            if isnan(obj1):
                raise ValueError("Error! Objective function value NaN!")
            if self.stopper.run(problem, iiter, verbose=verbose):
                break

        # Writing last inverted model
        self.save_results(iiter, problem, force_save=True, force_write=True)
        if self.create_msg:
            msg = 90 * "#" + "\n"
            msg += "PRECONDITIONED " if precond else ""
            msg += "LINEAR %s SOLVER FOR SYMMETRIC MATRIX log file end\n" \
                   % ("STEEPEST-DESCENT" if self.steepest else "CONJUGATE GRADIENT")
            msg += 90 * "#" + "\n"
            if verbose:
                print(msg.replace("log file ", ""))
            if self.logger:
                self.logger.addToLog(msg)
        # Clear restart object
        self.restart.clear_restart()



class Bounds:
    """Class used to enforce boundary constraints during the inversion"""

    def __init__(self, minBound=None, maxBound=None):
        """
        Bounds constructor
        minBound    = [None] - vector class; vector containing minimum values of the model vector
        maxBound    = [None] - vector class; vector containing maximum values of the model vector
        """
        self.minBound = minBound
        self.maxBound = maxBound
        if minBound is not None:
            self.minBound = minBound.clone()
        if maxBound is not None:
            self.maxBound = maxBound.clone()
        # If only the lower bound was provided we use the opposite of the lower bound to clip the values
        if self.minBound is not None and self.maxBound is None:
            self.minBound.scale(-1.0)
        return

    def apply(self, input_vec):
        """
        Function for applying the model bounds
        """
        if self.minBound is not None and self.maxBound is None:
            if not input_vec.checkSame(self.minBound):
                raise ValueError("Input vector not consistent with bound space")
            input_vec.scale(-1.0)
            input_vec.clipVector(input_vec, self.minBound)
            input_vec.scale(-1.0)
        elif self.minBound is None and self.maxBound is not None:
            if not input_vec.checkSame(self.maxBound):
                raise ValueError("Input vector not consistent with bound space")
            input_vec.clipVector(input_vec, self.maxBound)
        elif self.minBound is not None and self.maxBound is not None:
            if (not (input_vec.checkSame(self.minBound) and input_vec.checkSame(
                    self.maxBound))):
                raise ValueError("Input vector not consistent with bound space")
            input_vec.clipVector(self.minBound, self.maxBound)
        return


class Problem:
    """Problem parent object"""

    # Default class methods/functions
    def __init__(self, minBound=None, maxBound=None, boundProj=None):
        """Default class constructor for Problem"""
        if minBound is not None or maxBound is not None:
            # Simple box bounds
            self.bounds = Bounds(minBound, maxBound)  # Setting the bounds of the problem (if necessary)
        elif boundProj is not None:
            # Projection operator onto the bounds
            self.bounds = boundProj
        # Setting common variables
        self.obj_updated = False
        self.res_updated = False
        self.grad_updated = False
        self.dres_updated = False
        self.fevals = 0
        self.gevals = 0
        self.counter = 0
        self.linear = False  # By default all problem are non-linear

    def __del__(self):
        """Default destructor"""
        return

    def setDefaults(self):
        """Default common variables for any inverse problem"""
        self.obj_updated = False
        self.res_updated = False
        self.grad_updated = False
        self.dres_updated = False
        self.fevals = 0
        self.gevals = 0
        self.counter = 0
        return

    def set_model(self, model):
        """Setting internal model vector"""
        if model.isDifferent(self.model):
            self.model.copy(model)
            self.obj_updated = False
            self.res_updated = False
            self.grad_updated = False
            self.dres_updated = False

    def set_residual(self, residual):
        """Setting internal residual vector"""
        # Useful for linear inversion (to avoid residual computation)
        if self.res.isDifferent(residual):
            self.res.copy(residual)
            # If residuals have changed, recompute gradient and objective function value
            self.grad_updated = False
            self.obj_updated = False
        self.res_updated = True
        return

    def get_model(self):
        """Accessor for model vector"""
        return self.model

    def get_dmodel(self):
        """Accessor for model vector"""
        return self.dmodel

    def get_rnorm(self, model):
        """Accessor for residual vector norm"""
        self.get_res(model)
        return self.get_res(model).norm()

    def get_gnorm(self, model):
        """Accessor for gradient vector norm"""
        return self.get_grad(model).norm()

    def get_obj(self, model):
        """Accessor for objective function"""
        self.set_model(model)
        if not self.obj_updated:
            self.res = self.get_res(self.model)
            self.obj = self.objf(self.res)
            self.obj_updated = True
        return self.obj

    def get_res(self, model):
        """Accessor for residual vector"""
        self.set_model(model)
        if not self.res_updated:
            self.fevals += 1
            self.res = self.resf(self.model)
            self.res_updated = True
        return self.res

    def get_grad(self, model):
        """Accessor for gradient vector"""
        self.set_model(model)
        if not self.grad_updated:
            self.res = self.get_res(self.model)
            self.grad = self.gradf(self.model, self.res)
            self.gevals += 1
            if self.linear:
                self.fevals += 1
            self.grad_updated = True
        return self.grad

    def get_dres(self, model, dmodel):
        """Accessor for dresidual vector (i.e., application of the Jacobian to Dmodel vector)"""
        self.set_model(model)
        if not self.dres_updated or dmodel.isDifferent(self.dmodel):
            self.dmodel.copy(dmodel)
            self.dres = self.dresf(self.model, self.dmodel)
            if self.linear:
                self.fevals += 1
            self.dres_updated = True
        return self.dres

    def get_fevals(self):
        """Accessor for number of objective function evalutions"""
        return self.fevals

    def get_gevals(self):
        """Accessor for number of gradient evalutions"""
        return self.gevals

    def objf(self, residual):
        """Dummy objf running method, must be overridden in the derived class"""
        raise NotImplementedError("Implement objf for problem in the derived class!")

    def resf(self, model):
        """Dummy resf running method, must be overridden in the derived class"""
        raise NotImplementedError("Implement resf for problem in the derived class!")

    def dresf(self, model, dmodel):
        """Dummy dresf running method, must be overridden in the derived class"""
        raise NotImplementedError("Implement dresf for problem in the derived class!")

    def gradf(self, model, residual):
        """Dummy gradf running method, must be overridden in the derived class"""
        raise NotImplementedError("Implement gradf for problem in the derived class!")



# Module containing the definition of inverse problems where the ADMM method is used
from math import isnan
from occamypy import Operator, Vector,
import pyOperator as pyOp
import pyVector as pyVec
from pyLinearSolver import LCGsolver, LSQRsolver
from pySparseSolver import ISTAsolver
from pyProblem import Problem, ProblemL1Lasso, ProblemL2LinearReg, ProblemL2Linear, ProblemLinearReg
from pySolver import Solver
from pySparseSolver import *
from pyStopper import BasicStopper
from sys_util import logger


class ADMM(Solver):
    """Alternate Directions of Multipliers Method (ADMM) for GeneralizedLasso problems"""
    
    # Default class methods/functions
    def __init__(self, stopper, logger=None, niter_linear=5, niter_lasso=5, rho=None,
                 rho_auto=True, rho_ratio=10., rho_weight=2., warm_start=False, linear_solver='CG'):
        """
        Constructor for ADMM Solver
        :param stopper       : stopper object
        :param logger        : logger object
        :param niter_linear  : int; number of iterations for solving the linear problem [5]
        :param niter_lasso   : int; number of iterations for solving the lasso problem [5]
        :param rho           : float; penalty parameter rho (if None it is initialized as 2*gamma+.1)
        :param rho_auto      : bool; update rho automatically
        :param rho_ratio     : float; norm ratio between residuals for updating rho [10]
        :param rho_weight    : float; scaling factor for updating rho [2]
        :param warm_start    : bool; linear solver uses previous solution [False]
        :param linear_solver : str; linear solver to be used [CG, SD, LSQR]
        """
        # Calling parent construction
        super(ADMM, self).__init__()
        
        self.stopper = stopper
        self.logger = logger
        self.stopper.logger = self.logger
        self.niter_linear = niter_linear
        self.niter_lasso = niter_lasso
        self.warm_start = warm_start
        
        if linear_solver == 'CG':
            self.solver_linear = LCGsolver(BasicStopper(niter=self.niter_linear), steepest=False, logger=self.logger)
        elif linear_solver == 'SD':
            self.solver_linear = LCGsolver(BasicStopper(niter=self.niter_linear), steepest=True, logger=self.logger)
        elif linear_solver == 'LSQR':
            self.solver_linear = LSQRsolver(BasicStopper(niter=self.niter_linear), logger=self.logger)
        else:
            raise ValueError('ERROR! Solver has to be CG, SD or LSQR')
        
        self.solver_lasso = ISTAsolver(BasicStopper(niter=self.niter_lasso), fast=True, logger=self.logger)
        
        self.rho = rho  # ADMM penalty parameter
        self.rho_ratio = rho_ratio  # norm ratio between residuals for updating rho
        self.rho_weight = rho_weight  # scaling factor for updating rho
        self.rho_auto = rho_auto  # whether to update rho
        self.primal = None  # primal residual vector (r = A x + B z - c)
        self.dual = None  # dual residual vector (s = rho A.H B r)
        
        # print formatting
        self.iter_msg = "iter = %s, obj = %.5e, df_obj = %.2e, reg_obj = %.2e, resnorm = %.2e"
    
    def __del__(self):
        print('Destructor called, ADMM deleted')
    
    def update_rho(self):
        """update penalty parameter rho as suggested in boyd2010distributed (3.13)"""
        if self.primal.norm(2) > self.rho_ratio * self.dual.norm(2):
            self.rho = self.rho * self.rho_weight
        elif self.dual.norm(2) > self.rho_ratio * self.primal.norm(2):
            self.rho = self.rho / self.rho_weight
        else:
            pass
    
    def init_rho(self, gamma):
        self.rho = 2 * gamma + .1
    
    def run(self, problem, verbose=False, inner_verbose=False, restart=False, initial_guess=None):
        
        assert type(problem) == ProblemLinearReg, 'problem has to be a ProblemLinearReg'
        if problem.nregsL1 == 0:
            raise ValueError('ERROR! Provide at least one L1 regularizer!')
        
        self.create_msg = verbose or self.logger
        
        # I want to set dfw=1, so:
        gamma = max(problem.epsL1)
        
        # A is the Vstack of the L1 reg operators, scaled by their respective eps
        # we divide by gamma as gamma becomes the lambda value for the FISTA problem
        A = problem.regL1_op * [e / gamma for e in problem.epsL1]
        
        # initialize all others variables
        if self.rho is None:
            self.init_rho(gamma)
        
        admm_mdl = problem.model.clone().zero() if initial_guess is None else initial_guess.clone()
        y = A.range.clone().zero()
        u = y.clone()
        self.dual = A.domain.clone().zero()
        
        if restart:
            self.restart.read_restart()
            outer_iter = self.restart.retrieve_parameter("iter")
            initial_obj_value = self.restart.retrieve_parameter("obj_initial")
            admm_mdl = self.restart.retrieve_vector("admm_mdl")
            if self.create_msg:
                msg = "Restarting previous solver run from: %s" % self.restart.restart_folder
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog(msg)
        else:
            outer_iter = 0
            if self.create_msg:
                msg = 90 * '#' + '\n'
                msg += "\t\t\t\t\tADMM ALGORITHM log file\n\n"
                msg += "\tRestart folder: %s\n" % self.restart.restart_folder
                msg += "\tModeling Operator:\t\t%s\n" % problem.op
                if problem.nregsL2 != 0:
                    msg += "\tL2 Regularizer ops:\t\t" + ", ".join(["%s" % op for op in problem.regL2_op.ops]) + "\n"
                    msg += "\tL2 Regularizer weights:\t" + ", ".join(["{:.2e}".format(e) for e in problem.epsL2]) + "\n"
                msg += "\tL1 Regularizer ops:\t\t" + ", ".join(["%s" % op for op in problem.regL1_op.ops]) + "\n"
                msg += "\tL1 Regularizer weights:\t" + ", ".join(["{:.2e}".format(e) for e in problem.epsL1]) + "\n"
                msg += "\tPenalty parameter:\t\t%.2e\n" % self.rho
                msg += 90 * '#' + '\n'
                if verbose:
                    print(msg.replace(" log file", ""))
                if self.logger:
                    self.logger.addToLog(msg)
        
        # Main iteration loop
        while True:
            obj0 = problem.get_obj(admm_mdl)
            
            if outer_iter == 0:
                initial_obj_value = obj0
                self.restart.save_parameter("obj_initial", initial_obj_value)
                if self.create_msg:
                    msg = self.iter_msg % (str(outer_iter).zfill(self.stopper.zfill),
                                           obj0,
                                           problem.obj_terms[0],
                                           obj0 - problem.obj_terms[0],
                                           problem.get_rnorm(admm_mdl))
                    if verbose:
                        print(msg)
                    if self.logger:
                        self.logger.addToLog("\n" + msg)
                
                if isnan(obj0):
                    raise ValueError("Objective function values NaN!")
            
            if obj0 == 0:
                print("Objective function is 0!")
                break
            
            self.save_results(outer_iter, problem, force_save=False)
            
            # 1) update x
            # Linear Problem:       1/2 | Op x - d| + epsL2   | R2 x -  dr  |
            #                                         rho/2   | A  x - (y-u)|
            regL2_op_scaled_list = [problem.epsL2[i] * problem.regL2_op.ops[i] for i in range(problem.nregsL2)]
            regA_op_scaled_list = [self.rho / 2 * A.ops[i] for i in range(A.n)]
            reg_op = pyOp.Vstack(
                pyOp.Vstack(regL2_op_scaled_list) if len(regL2_op_scaled_list) != 0 else None,
                pyOp.Vstack(regA_op_scaled_list) if len(regA_op_scaled_list) != 0 else None,
            )
            prior = pyVec.superVector(problem.dataregsL2, y.clone().scaleAdd(u, 1., -1.))
            linear_problem = ProblemL2LinearReg(
                model=admm_mdl.clone().zero() if not self.warm_start else admm_mdl.clone(),
                data=problem.data,
                op=problem.op,
                reg_op=reg_op,  # pyOp.Vstack(problem.regL2_op, A),
                epsilon=1.,  # problem.epsL2 + [self.rho / 2] * A.n,
                prior_model=prior,
                minBound=problem.minBound, maxBound=problem.maxBound, boundProj=problem.boundProj
            )
            if outer_iter == 0 and initial_guess is not None:
                linear_problem.model = initial_guess.clone()
            
            self.solver_linear.run(linear_problem, verbose=inner_verbose)
            
            admm_mdl = linear_problem.model.clone()
            
            # 2) update y
            # lasso problem: rho/2 | A x - z + u|_2^2 + gamma | y |_1
            # this means to solve: 1/2 | I y - (Ax + u)| + gamma/rho | y |_1
            Ax = A * admm_mdl
            lasso_problem = ProblemL1Lasso(  # TODO it stops at the second iteration
                model=y.clone().zero(),
                data=Ax.clone() + u,
                op=pyOp.Identity(y),
                op_norm=1.,
                lambda_value=gamma / self.rho,
                minBound=problem.minBound, maxBound=problem.maxBound, boundProj=problem.boundProj
            )
            self.solver_lasso.setDefaults()
            self.solver_lasso.run(lasso_problem, verbose=inner_verbose)
            y = lasso_problem.model.clone()
            
            # 3) update penalty parameter and scaled dual variable
            self.primal = Ax.clone() - y
            
            A.adjoint(False, self.dual, self.primal)
            self.dual.scale(-self.rho)
            self.update_rho()
            u.__add__(self.primal).scale(1 / self.rho)  # u = (u + r)/rho
            
            outer_iter += 1
            # check objective function
            obj1 = problem.get_obj(admm_mdl)
            # if obj1 >= obj0:
            #     msg = "Objective function didn't reduce, will terminate solver:\n\t" \
            #           "obj_new = %.2e\tobj_cur = %.2e" % (obj1, obj0)
            #     if verbose:
            #         print(msg)
            #     if self.logger:
            #         self.logger.addToLog(msg)
            #     break
            
            # iteration info
            if self.create_msg:
                msg = self.iter_msg % (str(outer_iter).zfill(self.stopper.zfill),
                                       obj1,
                                       problem.obj_terms[0],
                                       obj1 - problem.obj_terms[0],
                                       problem.get_rnorm(admm_mdl))
                if verbose:
                    print(msg)
                if self.logger:
                    self.logger.addToLog("\n" + msg)
            
            # saving in case of restart
            self.restart.save_parameter("iter", outer_iter)
            self.restart.save_vector("admm_mdl", admm_mdl)
            
            if self.stopper.run(problem, outer_iter, initial_obj_value, verbose):
                break
        
        # writing last inverted model
        self.save_results(outer_iter, problem, model=None, force_save=False, force_write=False)
        
        # ending message and log file
        if self.create_msg:
            msg = 90 * '#' + '\n'
            msg += "\t\t\t\t\tADMM ALGORITHM log file end\n"
            msg += 90 * '#'
            if verbose:
                print(msg.replace(" log file", ""))
            if self.logger:
                self.logger.addToLog("\n" + msg)
        
        # Clear restart object
        self.restart.clear_restart()


def main():
    from sys import path
    path.insert(0, '.')
    import numpy as np
    import matplotlib.pyplot as plt
    # plt.style.use('ggplot')
    import pyNpOperator
    from pyProblem import ProblemLinearReg
    from pySparseSolver import SplitBregmanSolver
    
    PLOT = True
    EXAMPLE = 'monarch'  # must be noisy, gaussian1D, gaussian2D, medical or monarch
    
    if EXAMPLE == 'noisy':
        # data examples
        np.random.seed(1)
        nx = 101
        x = pyVec.vectorIC((nx,)).zero()
        x.getNdArray()[:nx // 2] = 10
        x.getNdArray()[nx // 2:3 * nx // 4] = -5
        
        Iop = pyOp.Identity(x)
        TV = pyNpOperator.FirstDerivative(x)
        L = pyNpOperator.SecondDerivative(x)
        
        n = x.clone()
        n.getNdArray()[:] = np.random.normal(0, 1.0, nx)
        y = Iop * (x.clone() + n)
        
        derivative = TV * x
        
        # if PLOT:
        #     plt.figure(figsize=(5, 4))
        #     plt.plot(x.getNdArray(), 'k', lw=1, label='x')
        #     plt.plot(y.getNdArray(), '.k', label='y=x+n')
        #     plt.plot(derivative.getNdArray(), '.b', lw=2, label='∂x')
        #     plt.legend()
        #     plt.title('Model, Data and Derivative')
        #     plt.show()
        #
        # # CG solver
        # problemLS = ProblemL2Linear(x.clone().zero(), y, Iop)
        # CG = LCGsolver(BasicStopper(niter=30))
        # CG.run(problemLS, verbose=True)
        # if PLOT:
        #     plt.figure(figsize=(5, 4))
        #     plt.plot(x.getNdArray(), 'k', lw=1, label='x')
        #     plt.plot(y.getNdArray(), '.k', label='y=x+n')
        #     plt.plot(problemLS.model.getNdArray(), 'r', lw=2, label='x_inv')
        #     plt.legend()
        #     plt.title('Least-Squares CG')
        #     plt.show()
        #
        # # LSQR solver
        # problemLSQR = ProblemL2Linear(x.clone().zero(), y, Iop)
        # LSQR = LSQRsolver(BasicStopper(niter=30))
        # LSQR.run(problemLSQR, verbose=True)
        # if PLOT:
        #     plt.figure(figsize=(5, 4))
        #     plt.plot(x.getNdArray(), 'k', lw=1, label='x')
        #     plt.plot(y.getNdArray(), '.k', label='y=x+n')
        #     plt.plot(problemLSQR.model.getNdArray(), 'r', lw=2, label='x_inv')
        #     plt.legend()
        #     plt.title('Least-Squares LSQR')
        #     plt.show()
        #
        # # CG solver with L2 regularization
        # problemLSR = ProblemL2LinearReg(x.clone().zero(), y, Iop, np.sqrt(50), L)
        # CG = LCGsolver(BasicStopper(niter=30))
        # CG.run(problemLSR, verbose=True)
        # if PLOT:
        #     plt.figure(figsize=(5, 4))
        #     plt.plot(x.getNdArray(), 'k', lw=1, label='x')
        #     plt.plot(y.getNdArray(), '.k', label='y=x+n')
        #     plt.plot(problemLSR.model.getNdArray(), 'r', lw=2, label='x_inv')
        #     plt.legend()
        #     plt.title('CG with Laplacian reg')
        #     plt.show()
        #
        # # LSQR solver with L2 regularization
        # problemLSR_1 = ProblemL2LinearReg(x.clone().zero(), y, Iop, np.sqrt(50), L)
        # LSQR = LSQRsolver(BasicStopper(niter=30))
        # LSQR.run(problemLSR_1, verbose=True)
        # if PLOT:
        #     plt.figure(figsize=(5, 4))
        #     plt.plot(x.getNdArray(), 'k', lw=1, label='x')
        #     plt.plot(y.getNdArray(), '.k', label='y=x+n')
        #     plt.plot(problemLSR_1.model.getNdArray(), 'r', lw=2, label='x_inv')
        #     plt.legend()
        #     plt.title('LSQR with Laplacian reg')
        #     plt.show()
        #
        # # FISTA
        # problemFISTA = ProblemL1Lasso(x.clone().zero(), y, Iop, lambda_value=1, op_norm=1)
        # FISTA = ISTAsolver(BasicStopper(niter=300), fast=True)
        # FISTA.run(problemFISTA, verbose=True)
        # if PLOT:
        #     plt.figure(figsize=(5, 4))
        #     plt.plot(x.getNdArray(), 'k', lw=1, label='x')
        #     plt.plot(y.getNdArray(), '.k', label='y=x+n')
        #     plt.plot(problemFISTA.model.getNdArray(), 'r', lw=2, label='x_inv')
        #     plt.legend()
        #     plt.title('FISTA inversion')
        #     plt.show()
        
        # # SplitBregman
        problemSB = ProblemLinearReg(x.clone().zero(), y, Iop, regsL1=TV, epsL1=3.0)
        SB = SplitBregmanSolver(BasicStopper(niter=100), lambd=0.03, niter_inner=1, niter_solver=20,
                                linear_solver='LSQR', breg_weight=1.,
                                warm_start=True)  # , logger=logger("test_SB1.txt"))
        SB.setDefaults(save_obj=True)
        SB.run(problemSB, verbose=True, inner_verbose=False)
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.plot(x.getNdArray(), 'k', lw=1, label='x')
            plt.plot(y.getNdArray(), '.k', label='y=x+n')
            plt.plot(derivative.getNdArray(), ':k', lw=1, label='∂x')
            plt.plot(problemSB.model.getNdArray(), 'r', lw=2, label='x_inv')
            plt.plot((TV * problemSB.model).getNdArray(), ':r', lw=2, label='∂(x_inv)')
            plt.legend()
            plt.title('SB inversion')
            plt.show()
            # Objective function convergence
            plt.figure(figsize=(5, 4))
            plt.plot(np.log10(SB.obj / SB.obj[0]), 'r', lw=1, label='SplitBregman')
            obj_true = problemSB.get_obj(x)
            plt.plot([np.log10(obj_true / SB.obj[0])] * len(SB.obj), 'k--', lw=1, label='true solution obj value')
            plt.legend()
            plt.title('Convergence curve')
            plt.show()
        
        # ADMM
        # problemADMM = ProblemLinearReg(x.clone().zero(), y, Iop, regsL1=TV, epsL1=3.)
        #
        # ADMM = ADMMsolver(BasicStopper(niter=30), niter_linear=10, niter_lasso=50)
        # ADMM.run(problemADMM, verbose=True, inner_verbose=False)
        # if PLOT:
        #     plt.figure(figsize=(5, 4))
        #     plt.plot(x.getNdArray(), 'k', lw=1, label='x')
        #     plt.plot(y.getNdArray(), '.k', label='y=x+n')
        #     plt.plot(derivative.getNdArray(), ':k', lw=1, label='∂x')
        #     plt.plot(problemADMM.model.getNdArray(), 'r', lw=2, label='x_inv')
        #     plt.plot((TV * problemADMM.model).getNdArray(), ':r', lw=2, label='∂(x_inv)')
        #     plt.legend()
        #     plt.title('ADMM inversion')
        #     plt.show()
    
    if EXAMPLE == 'gaussian1D':
        # data examples
        np.random.seed(1)
        nx = 101
        x = pyVec.vectorIC((nx,)).zero()
        x.getNdArray()[:nx // 2] = 10
        x.getNdArray()[nx // 2:3 * nx // 4] = -5
        
        Op = pyNpOperator.GaussianFilter(x, sigma=(5))
        TV = pyNpOperator.FirstDerivative(x)
        L = pyNpOperator.SecondDerivative(x)
        y = Op * x
        dx = TV * x
        
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.plot(x.getNdArray(), 'k', lw=1, label='x')
            plt.plot(y.getNdArray(), 'b', label='y=Gx')
            plt.plot(dx.getNdArray(), '.k', lw=2, label='∂x')
            plt.legend()
            plt.title('Model, Data and Derivative')
            plt.show()
        
        # # CG solver
        problemLS = ProblemL2Linear(x.clone().zero(), y, Op)
        CG = LCGsolver(BasicStopper(niter=500))
        CG.run(problemLS)
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.plot(x.getNdArray(), 'k', label='x')
            plt.plot(y.getNdArray(), 'b', label='y=Gx')
            plt.plot(problemLS.model.getNdArray(), 'r', label='x_inv')
            plt.legend()
            plt.title('Least-Squares CG')
            plt.show()
        
        # LSQR solver
        problemLSQR = ProblemL2Linear(x.clone().zero(), y, Op)
        LSQR = LSQRsolver(BasicStopper(niter=500))
        LSQR.run(problemLSQR, verbose=True)
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.plot(x.getNdArray(), 'k', label='x')
            plt.plot(y.getNdArray(), 'b', label='y=Gx')
            plt.plot(problemLSQR.model.getNdArray(), 'r', label='x_inv')
            plt.legend()
            plt.title('Least-Squares LSQR')
            plt.show()
        #
        # CG solver with L2 regularization
        problemLSR = ProblemL2LinearReg(x.clone().zero(), y, Op, 1, L)
        CG = LCGsolver(BasicStopper(niter=500))
        CG.run(problemLSR, verbose=True)
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.plot(x.getNdArray(), 'k', label='x')
            plt.plot(y.getNdArray(), 'b', label='y=Gx')
            plt.plot(problemLSR.model.getNdArray(), 'r', label='x_inv')
            plt.legend()
            plt.title('CG with Laplacian reg')
            plt.show()
        #
        # # LSQR solver with L2 regularization
        # problemLSR_1 = ProblemL2LinearReg(x.clone().zero(), y, Iop, np.sqrt(50), L)
        # LSQR = LSQRsolver(BasicStopper(niter=30))
        # LSQR.run(problemLSR_1, verbose=True)
        # if PLOT:
        #     plt.figure(figsize=(5, 4))
        #     plt.plot(x.getNdArray(), 'k', lw=1, label='x')
        #     plt.plot(y.getNdArray(), '.k', label='y=x+n')
        #     plt.plot(problemLSR_1.model.getNdArray(), 'r', lw=2, label='x_inv')
        #     plt.legend()
        #     plt.title('LSQR with Laplacian reg')
        #     plt.show()
        #
        # FISTA
        problemFISTA = ProblemL1Lasso(x.clone().zero(), y, Op, lambda_value=.5, op_norm=1.05)
        FISTA = ISTAsolver(BasicStopper(niter=300), fast=True)
        FISTA.run(problemFISTA, verbose=True)
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.plot(x.getNdArray(), 'k', label='x')
            plt.plot(y.getNdArray(), 'b', label='y=Gx')
            plt.plot(problemFISTA.model.getNdArray(), 'r', label='x_inv')
            plt.legend()
            plt.title('FISTA inversion')
            plt.show()
        
        # # SplitBregman
        problemSB = ProblemLinearReg(x.clone().zero(), y, Op, regsL1=TV, epsL1=10.0)
        SB = SplitBregmanSolver(BasicStopper(niter=100), lambd=0.03, niter_inner=1, niter_solver=20,
                                linear_solver='CG', breg_weight=1.,
                                warm_start=False)  # , logger=logger("test_SB1.txt"))
        SB.run(problemSB, verbose=True, inner_verbose=False)
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.plot(x.getNdArray(), 'k', label='x')
            plt.plot(y.getNdArray(), 'b', label='y=Gx')
            plt.plot(dx.getNdArray(), ':k', lw=1, label='∂x')
            plt.plot(problemSB.model.getNdArray(), 'r', label='x_inv')
            plt.plot((TV * problemSB.model).getNdArray(), ':r', lw=1, label='∂(x_inv)')
            plt.legend()
            plt.title('SB inversion')
            plt.show()
            # Objective function convergence
            plt.figure(figsize=(5, 4))
            plt.plot(np.log10(SB.obj / SB.obj[0]), 'r', lw=1, label='SplitBregman')
            obj_true = problemSB.get_obj(x)
            plt.plot([np.log10(obj_true / SB.obj[0])] * len(SB.obj), 'k--', lw=1, label='true solution obj value')
            plt.legend()
            plt.title('Convergence curve')
            plt.show()
        
        # ADMM
        problemADMM = ProblemLinearReg(x.clone().zero(), y, Iop, regsL1=TV, epsL1=3.)
        
        ADMM = ADMM(BasicStopper(niter=30), niter_linear=10, niter_lasso=50)
        ADMM.run(problemADMM, verbose=True, inner_verbose=False)
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.plot(x.getNdArray(), 'k', lw=1, label='x')
            plt.plot(y.getNdArray(), '.k', label='y=x+n')
            plt.plot(derivative.getNdArray(), ':k', lw=1, label='∂x')
            plt.plot(problemADMM.model.getNdArray(), 'r', lw=2, label='x_inv')
            plt.plot((TV * problemADMM.model).getNdArray(), ':r', lw=2, label='∂(x_inv)')
            plt.legend()
            plt.title('ADMM inversion')
            plt.show()
    
    elif EXAMPLE == 'gaussian2D':
        x = pyVec.vectorIC(np.empty((301, 601))).set(0)
        x.getNdArray()[150, 300] = 1.0
        # x.getNdArray()[100, 200] = -5.0
        # x.getNdArray()[280, 400] = 1.0
        if PLOT:
            plt.figure(figsize=(6, 3))
            plt.imshow(x.getNdArray()), plt.colorbar()
            plt.title('Model')
            plt.show()
        
        G = pyNpOperator.GaussianFilter(x, [25, 15])
        y = G * x
        # y.scale(1./y.norm())
        if PLOT:
            plt.figure(figsize=(6, 3))
            plt.imshow(y.getNdArray()), plt.colorbar()
            plt.title('Data')
            plt.show()
        
        # CG solver
        problemLS = ProblemL2Linear(x.clone().zero(), y, G)
        CG = LCGsolver(BasicStopper(niter=30))
        CG.setDefaults()
        CG.run(problemLS, verbose=True)
        if PLOT:
            plt.figure(figsize=(6, 3))
            plt.imshow(problemLS.model.getNdArray()), plt.colorbar()
            plt.title('CG, %d its' % CG.stopper.niter)
            plt.show()
        
        # FISTA
        problemFISTA = ProblemL1Lasso(x.clone().zero(), y, G, lambda_value=1000, op_norm=1.)
        FISTA = ISTAsolver(BasicStopper(niter=1000), fast=True)
        FISTA.setDefaults()
        FISTA.run(problemFISTA, verbose=True)
        if PLOT:
            plt.figure(figsize=(6, 3))
            plt.imshow(problemFISTA.model.getNdArray()), plt.colorbar()
            plt.title(r'FISTA, $\lambda$=%.2e, %d its' % (problemFISTA.lambda_value, FISTA.stopper.niter))
            plt.show()
        
        # SplitBregman
        I = pyOp.Identity(x)
        problemSB = ProblemLinearReg(x.clone().zero(), y, G, regsL1=I, epsL1=10.)
        
        SB = SplitBregmanSolver(BasicStopper(niter=50), niter_inner=3, niter_solver=30,
                                linear_solver='LSQR', breg_weight=1., use_prev_sol=False)
        SB.setDefaults()
        SB.run(problemSB, verbose=True, inner_verbose=False)
        if PLOT:
            plt.figure(figsize=(6, 3))
            plt.imshow(problemSB.model.getNdArray()), plt.colorbar()
            plt.title('SplitBregman')
            plt.show()
        
        # ADMM
        problemADMM = ProblemLinearReg(x.clone().zero(), y, G, regsL1=I, epsL1=2.)
        
        ADMM = ADMM(BasicStopper(niter=10), niter_linear=30, niter_lasso=100)
        ADMM.setDefaults()
        ADMM.run(problemADMM, verbose=True, inner_verbose=True)
        if PLOT:
            plt.figure(figsize=(6, 3))
            plt.imshow(problemADMM.model.getNdArray()), plt.colorbar()
            plt.title('ADMM')
            plt.show()
    
    elif EXAMPLE == 'medical':
        x = pyVec.vectorIC(np.load('../testdata/shepp_logan_phantom.npy', allow_pickle=True).astype(np.float32))
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.imshow(x.getNdArray(), cmap='bone', vmin=x.min(), vmax=x.max()), plt.colorbar()
            plt.title('Model')
            plt.show()
        
        # nh = [5, 10]
        # hz = np.exp(-0.1 * np.linspace(-(nh[0] // 2), nh[0] // 2, nh[0]) ** 2)
        # hx = np.exp(-0.03 * np.linspace(-(nh[1] // 2), nh[1] // 2, nh[1]) ** 2)
        # hz /= np.trapz(hz)  # normalize the integral to 1
        # hx /= np.trapz(hx)  # normalize the integral to 1
        # h = hz[:, np.newaxis] * hx[np.newaxis, :]
        # if PLOT:
        #     plt.figure(figsize=(5, 4))
        #     plt.imshow(h, aspect='equal'), plt.colorbar()
        #     plt.title('Blurring Kernel')
        #     plt.show()
        # Blurring = pyNpOperator.ConvNDscipy(model=x, kernel=pyVec.vectorIC(h))
        Blurring = pyNpOperator.GaussianFilter(x, [3, 5])
        
        y = Blurring * x
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.imshow(y.getNdArray(), cmap='bone', vmin=x.min(), vmax=x.max()), plt.colorbar()
            plt.title('Data')
            plt.show()
        
        # CG solver
        problemLS = ProblemL2Linear(x.clone().zero(), y, Blurring, minBound=x.clone().set(0.0))
        CG = LCGsolver(BasicStopper(niter=400))
        CG.run(problemLS, verbose=True)
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.imshow(problemLS.model.getNdArray(), cmap='bone', vmin=x.min(), vmax=x.max()), plt.colorbar()
            plt.title('CG, %d iter' % CG.stopper.niter)
            plt.show()
        
        # FISTA
        # problemFISTA = ProblemL1Lasso(x.clone().zero(), y, Blurring, lambda_value=1, op_norm=1.25)
        # FISTA = ISTAsolver(BasicStopper(niter=100), fast=True)
        # FISTA.run(problemFISTA, verbose=True)
        # if PLOT:
        #     plt.figure(figsize=(5, 4))
        #     plt.imshow(problemFISTA.model.getNdArray(), cmap='bone'), plt.colorbar()
        #     plt.title(r'FISTA, $\lambda$=%.2e, %d iter'
        #               % (problemFISTA.lambda_value, FISTA.stopper.niter))
        #     plt.show()
        
        # SplitBregman
        # the gradient of the image is 6e3
        D = pyNpOperator.TotalVariation(x)
        I = pyOp.Identity(x)
        
        problemSB = ProblemLinearReg(x.clone().zero(), y, Blurring, regsL1=D, epsL1=0.00005,
                                     minBound=x.clone().set(0.0))
        
        SB = SplitBregmanSolver(BasicStopper(niter=300), lambd=0.1, niter_inner=1, niter_solver=30,
                                linear_solver='LSQR', breg_weight=1., warm_start=True)
        SB.setDefaults(save_obj=True)
        SB.run(problemSB, verbose=True, inner_verbose=False)
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.imshow(problemSB.model.getNdArray(), cmap='bone', vmin=x.min(), vmax=x.max()), plt.colorbar()
            plt.title(r'SB TV, $\varepsilon=%.2e$, %d iter'
                      % (problemSB.epsL1[0], SB.stopper.niter))
            plt.show()
            plt.figure(figsize=(5, 4))
            plt.plot(np.log10(SB.obj / SB.obj[0]), 'r', lw=1, label='SplitBregman')
            obj_true = problemSB.get_obj(x)
            plt.plot([np.log10(obj_true / SB.obj[0])] * len(SB.obj), 'k--', lw=1, label='true solution obj value')
            plt.legend()
            plt.title('Convergence curve')
            plt.show()
        
        # ADMM
        # problemADMM = ProblemLinearReg(x.clone().zero(), y, Blurring,
        #                              regsL1=D, epsL1=.1)
        #
        # ADMM = ADMMsolver(BasicStopper(niter=10), niter_linear=30, niter_lasso=10)
        # ADMM.setDefaults(save_obj=True, save_model=True)
        # ADMM.run(problemADMM, verbose=True, inner_verbose=True)
        # if PLOT:
        #     plt.figure(figsize=(5, 4))
        #     plt.imshow(problemADMM.model.getNdArray(), cmap='bone'), plt.colorbar()
        #     plt.title(r'ADMM TV, $\varepsilon=%.2e$, %d iter'
        #               % (problemADMM.epsL1[0], ADMM.stopper.niter))
        #     plt.show()
    
    elif EXAMPLE == 'monarch':
        x = pyVec.vectorIC(np.load('../testdata/monarch.npy', allow_pickle=True).astype(np.float32))
        np.random.seed(12345)
        sigma = 0.05
        y = x.clone()
        y.getNdArray()[:] += np.random.normal(0.0, sigma, y.shape)
        Op = pyOp.Identity(x)
        TV = pyNpOperator.TotalVariation(x)
        
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.imshow(x.arr, cmap='gray'), plt.colorbar()
            plt.title('Model')
            plt.show()
            
            plt.figure(figsize=(5, 4))
            plt.imshow(y.arr, cmap='gray'), plt.colorbar()
            plt.title('Data, std=%.2f' % sigma)
            plt.show()
        
        problemADMM = ProblemLinearReg(x.clone().zero(), y, Op, regsL1=TV, epsL1=.04)
        
        ADMM = ADMM(BasicStopper(niter=200), niter_linear=30, niter_lasso=10)
        ADMM.run(problemADMM, verbose=True, inner_verbose=False)
        if PLOT:
            plt.figure(figsize=(5, 4))
            plt.imshow(problemADMM.model.getNdArray(), cmap='gray'), plt.colorbar()
            plt.title(r'ADMM TV, $\varepsilon=%.2e$, %d iter'
                      % (problemADMM.epsL1[0], ADMM.stopper.niter))
            plt.show()
    
    else:
        raise ValueError("EXAMPLE has to be one of noisy, gaussian, medical or monarch")
    
    return 0


if __name__ == '__main__':
    main()



from .base import *
from .linear import *
from .nonlinear import *

__all__ = [
    "Bounds",
    "Problem",
    "LeastSquares",
    "LeastSquaresSymmetric",
    "LeastSquaresRegularized",
    "Lasso",
    "GeneralizedLasso",
    "NonlinearLeastSquares",
    "NonlinearLeastSquaresRegularized",
    "VarProRegularized",
]



from __future__ import division, print_function, absolute_import
import time
from copy import deepcopy
import numpy as np

from occamypy.vector import Vector, superVector
from occamypy import solver as S
from occamypy import problem as P


class Operator:
    """Abstract python operator class"""

    # Default class methods/functions
    def __init__(self, domain, range):
        """Generic class for operator"""
        self.domain = domain.cloneSpace()
        self.range = range.cloneSpace()

    def __del__(self):
        """Default destructor"""
        return

    def __str__(self):
        return "Operator"
    
    def __repr__(self):
        return self.__str__()

    # unary operators
    def __add__(self, other):  # self + other
        if isinstance(other, Operator):
            return _sumOperator(self, other)
        else:
            raise TypeError('Argument must be an Operator')

    def __sub__(self, other):  # self - other
        self.__add__(-other)

    def __neg__(self):  # -self
        return _scaledOperator(self, -1)

    def __mul__(self, other):  # self * other
        return self.dot(other)

    __rmul__ = __mul__  # other * self

    def __truediv__(self, other, niter=2000):
        """x = A / y through CG"""

        if not self.range.checkSame(other):
            raise ValueError('Operator range and data domain mismatch')

        stopper = S.BasicStopper(niter=niter)
        problem = P.LeastSquares(model=self.domain.clone(), data=other, op=self)
        CGsolver = S.CG(stopper)
        CGsolver.run(problem, verbose=False)

        return problem.model

    # main function for all kinds of multiplication
    def dot(self, other):
        """Matrix-matrix or matrix-vector or matrix-scalar multiplication."""
        if isinstance(other, Operator):  # A * B
            return _prodOperator(self, other)
        elif type(other) in [int, float]:  # A * c or c * A
            return _scaledOperator(self, other)
        elif isinstance(other, list) and isinstance(self, Vstack):
            assert len(other) == self.n, "Other lenght and self lenght mismatch"
            return Vstack([_scaledOperator(self.ops[i], other[i]) for i in range(self.n)])
        elif isinstance(other, list) and isinstance(self, Hstack):
            assert len(other) == self.n, "Other lenght and self lenght mismatch"
            return Hstack([_scaledOperator(self.ops[i], other[i]) for i in range(self.n)])
        elif isinstance(other, Vector) or isinstance(other, superVector):  # A * x
            temp = self.range.clone()
            self.forward(False, other, temp)
            return temp
        else:
            raise TypeError('Expected Operator, (super)Vector or scalar, got %r' % other)

    def getDomain(self):
        """Function to return operator domain"""
        return self.domain

    def getRange(self):
        """Function to return operator range"""
        return self.range

    def setDomainRange(self, domain, range):
        """Function to set (cloning space) domain and range of the operator"""
        self.domain = domain.cloneSpace()
        self.range = range.cloneSpace()
        return

    def checkDomainRange(self, x, y):
        """Function to check model and data vector sizes"""
        if not self.domain.checkSame(x):
            raise ValueError("Provided x vector does not match operator domain")
        if not self.range.checkSame(y):
            raise ValueError("Provided y vector does not match operator range")

    def powerMethod(self, verbose=False, tol=1e-8, niter=None, eval_min=False, return_vec=False):
        """
        Function to estimate maximum eigenvalue of the operator:

        :param return_vec: boolean - Return the estimated eigenvectors [False]
        :param niter: int - Maximum number of operator applications [None]
            if not provided, the function will continue until the tolerance is reached)
        :param eval_min: boolean - Compute the minimum eigenvalue [False]
        :param verbose: boolean - Print information to screen as the method is being run [False]
        :param tol: float - Tolerance on the change of the estimated eigenvalues [1e-6]
        """
        # Cloning input and output vectors
        if verbose:
            print('Running power method to estimate maximum eigenvalue (operator L2 norm)')
        x = self.domain.clone()
        # Checking if matrix is square
        square = False
        try:
            if self.domain.checkSame(self.range):
                square = True
        except:
            pass
        if not square:
            if verbose:
                print("Note: operator is not square, the eigenvalue is associated to A'A not A!")
            d_temp = self.range.clone()
        y = self.domain.clone()
        # randomize the input vector
        x.rand()
        x.scale(1.0 / x.norm())  # Normalizing the initial vector
        y.zero()
        iiter = 0
        eigen_old = 0.0  # Previous estimated eigenvalue
        # Estimating maximum eigenvalue
        if verbose:
            print("Starting iterative process for maximum eigenvalue")
        # Starting the power iteration loop
        while True:
            # Applying adjoint if forward not square
            if square:
                self.forward(False, x, y)  # y = A x
            else:
                self.forward(False, x, d_temp)  # d = A x
                self.adjoint(False, y, d_temp)  # y = A' d = A' A x

            # Estimating eigenvalue (Rayleigh quotient)
            eigen = x.dot(y)  # eigen_i = x' A x / (x'x = 1.0)
            # x = y
            x.copy(y)
            # Normalization of the operator
            x.scale(1.0 / x.norm())
            # Stopping criteria (first number of iterations and then tolerance)
            iiter += 1
            if verbose:
                print("Estimated maximum eigenvalue at iter %d: %.2e" % (iiter, eigen))
            if niter is not None:
                if iiter >= niter:
                    if verbose:
                        print("Maximum number of iteration reached! Stopping iterative process!")
                    break
            # Checking change on the eigenvalue estimated value
            if abs(eigen - eigen_old) < abs(tol * eigen_old):
                if verbose:
                    print("Tolerance value reached! Stopping iterative process!")
                break
            # eigen_(i-1) = eigen_i
            eigen_old = eigen
        if eval_min:
            x_max = x.clone()  # Cloning "maximum" eigenvector
            eigen_max = deepcopy(eigen)
            # Re-initialize variables
            x.rand()
            x.scale(1.0 / x.norm())  # Normalizing the initial vector
            y.zero()
            iiter = 0
            eigen = 0.0  # Current estimated eigenvalue
            eigen_old = 0.0  # Previous estimated eigenvalue
            # Estimating the minimum eigenvalue
            # Shifting all eigenvalues by maximum one (i.e., A_min = A-muI)
            if verbose:
                print("Starting iterative process for minimum eigenvalue")
            while True:
                # Applying adjoint if forward not square
                if not square:
                    self.forward(False, x, d_temp)  # d = A x
                    self.adjoint(False, y, d_temp)  # y = A' d = A' A x
                else:
                    self.forward(False, x, y)  # y = A x
                # y = Ax - mu*Ix
                y.scaleAdd(x, 1.0, -eigen_max)
                # Estimating eigenvalue (Rayleigh quotient)
                eigen = x.dot(y)  # eigen_i = x' A_min x / (x'x = 1.0)
                # x = y
                x.copy(y)
                # Normalization of the operator
                x.scale(1.0 / x.norm())
                # Stopping criteria (first number of iterations and then tolerance)
                iiter += 1
                if verbose:
                    print("Estimated minimum eigenvalue at iter %d: %.2e"
                          % (iiter, eigen + eigen_max))
                if niter is not None:
                    if iiter >= niter:
                        if verbose:
                            print("Maximum number of iteration reached! Stopping iterative process!")
                        break
                # Checking change on the eigenvalue estimated value
                if abs(eigen - eigen_old) < abs(tol * eigen_old):
                    if verbose:
                        print("Tolerance value reached! Stopping iterative process!")
                    break
                # eigen_(i-1) = eigen_i
                eigen_old = eigen
            x_min = x.clone()  # Cloning "minimum" eigenvector
            eigen_min = deepcopy(eigen + eigen_max)
            eigen = [eigen_max, eigen_min]
            x = [x_max, x_min]
        return (eigen, x) if return_vec else eigen

    def dotTest(self, verbose=False, tol=1e-4):
        """
        Function to perform dot-product tests.
        :param verbose  : boolean; Flag to print information to screen as the method is being run [False]
        :param tol      : float; The function throws a Warning if the relative error is greater than maxError [1e-4]
        """

        def _testing(add, dt1, dt2, tol, verbose=False):
            if isinstance(dt2, np.complex):
                dt2 = np.conj(dt2)
            err_abs = dt1 - dt2
            err_rel = err_abs / abs(dt2)
            if verbose:
                print("Dot products add=%s: domain=%.6e range=%.6e " % (str(add), abs(dt1), abs(dt2)))
                print("Absolute error: %.6e" % abs(err_abs))
                print("Relative error: %.6e \n" % abs(err_rel))
            if abs(err_rel) > tol:
                # # Deleting temporary vectors
                # del d1, d2, r1, r2
                raise Warning("\tDot products failure add=%s; relative error %.2e is greater than tolerance %.2e"
                              % (str(add), err_rel, tol))

        if verbose:
            msg = "Dot-product tests of forward and adjoint operators"
            print(msg+"\n"+"-"*len(msg))

        # Allocating temporary vectors for dot-product tests
        d1 = self.domain.clone()
        d2 = self.domain.clone()
        r1 = self.range.clone()
        r2 = self.range.clone()

        # Randomize the input vectors
        d1.rand()
        r1.rand()

        # Applying forward and adjoint operators with add=False
        if verbose:
            print("Applying forward operator add=False")
        start = time.time()
        self.forward(False, d1, r2)
        end = time.time()
        if verbose:
            print(" Runs in: %s seconds" % (end - start))
            print("Applying adjoint operator add=False")
        start = time.time()
        self.adjoint(False, d2, r1)
        end = time.time()
        if verbose:
            print(" Runs in: %s seconds" % (end - start))

        # Computing dot products
        dt1 = d1.dot(d2)
        dt2 = r1.dot(r2)
        _testing(False, dt1, dt2, tol, verbose)

        # Applying forward and adjoint operators with add=True
        if verbose:
            print("Applying forward operator add=True")
        start = time.time()
        self.forward(True, d1, r2)
        end = time.time()
        if verbose:
            print(" Runs in: %s seconds" % (end - start))
            print("Applying adjoint operator add=True")
        start = time.time()
        self.adjoint(True, d2, r1)
        end = time.time()
        if verbose:
            print(" Runs in: %s seconds" % (end - start))

        # Computing dot products
        dt1 = d1.dot(d2)
        dt2 = r1.dot(r2)
        _testing(True, dt1, dt2, tol, verbose)

        if verbose:
            print("-" * 49)

        # Deleting temporary vectors
        del d1, d2, r1, r2
        return

    def forward(self, add, model, data):
        """Forward operator"""
        raise NotImplementedError("Forward must be defined")

    def adjoint(self, add, model, data):
        """Adjoint operator"""
        raise NotImplementedError("Adjoint must be defined")

    def hermitian(self):
        """Instantiate the Hermitian operator"""
        return _Hermitian(self)

    H = property(hermitian)
    T = H  # misleading (H is the conjugate transpose), probably we can delete it


class _Hermitian(Operator):

    def __init__(self, op):
        super(_Hermitian, self).__init__(op.range, op.domain)
        self.op = op

    def forward(self, add, model, data):
        return self.op.adjoint(add, data, model)

    def adjoint(self, add, model, data):
        return self.op.forward(add, data, model)


class _CustomOperator(Operator):
    """Linear operator defined in terms of user-specified operations."""

    def __init__(self, domain, range, forward_function, adjoint_function):
        super(_CustomOperator, self).__init__(domain, range)
        self.forward_function = forward_function
        self.adjoint_function = adjoint_function

    def forward(self, add, model, data):
        return self.forward_function(add, model, data)

    def adjoint(self, add, model, data):
        return self.adjoint_function(add, model, data)


class Vstack(Operator):
    """
    Vertical stack of operators
        y1 = | A | x
        y2   | B |
    """
    
    def __init__(self, *args):
        """Constructor for the stacked operator"""
        
        self.ops = []
        for _, arg in enumerate(args):
            if arg is None:
                continue
            elif isinstance(arg, Operator):
                self.ops.append(arg)
            elif isinstance(arg, list):
                for op in arg:
                    if op is None:
                        continue
                    elif isinstance(op, Operator):
                        self.ops.append(op)
            else:
                raise TypeError('Argument must be either Operator or Vstack')
        
        # check range
        self.n = len(self.ops)
        op_range = []
        for idx in range(self.n):
            if idx < self.n - 1:
                if not self.ops[idx].domain.checkSame(self.ops[idx + 1].domain):
                    raise ValueError('Domain incompatibility between Op %d and Op %d' % (idx, idx + 1))
            op_range += [self.ops[idx].range]
        
        super(Vstack, self).__init__(domain=self.ops[0].domain, range=superVector(op_range))
    
    def __str__(self):
        return " VStack "
    
    def forward(self, add, model, data):
        """Forward operator Cm"""
        self.checkDomainRange(model, data)
        for idx in range(self.n):
            self.ops[idx].forward(add, model, data.vecs[idx])
    
    def adjoint(self, add, model, data):
        """Adjoint operator C'r = A'r1 + B'r2"""
        self.checkDomainRange(model, data)
        self.ops[0].adjoint(add, model, data.vecs[0])
        for idx in range(1, self.n):
            self.ops[idx].adjoint(True, model, data.vecs[idx])


class Hstack(Operator):
    """
    Horizontal stack of operators
        y = [A  B]  x1
                    x2
    """
    
    def __init__(self, *args):
        """Constructor for the stacked operator"""
        
        self.ops = []
        for _, arg in enumerate(args):
            if arg is None:
                continue
            elif isinstance(arg, Operator):
                self.ops.append(arg)
            elif isinstance(arg, list):
                for op in arg:
                    if op is None:
                        continue
                    elif isinstance(op, Operator):
                        self.ops.append(op)
            else:
                raise TypeError('Argument must be either Operator or Hstack')
        
        # check domain
        self.n = len(self.ops)
        domain = []
        for idx in range(self.n):
            if idx < self.n - 1:
                if not self.ops[idx].range.checkSame(self.ops[idx + 1].range):
                    raise ValueError('Range incompatibility between Op %d and Op %d' % (idx, idx + 1))
            domain += [self.ops[0].domain]
        super(Hstack, self).__init__(domain=superVector(domain), range=self.ops[0].range)
    
    def __str__(self):
        return " HStack "
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        self.ops[0].forward(add, model.vecs[0], data)
        for idx in range(1, self.n):
            self.ops[idx].forward(True, model.vecs[idx], data)
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        self.ops[0].adjoint(add, model.vecs[0], data)
        for idx in range(1, self.n):
            self.ops[idx].adjoint(add, model.vecs[idx], data)


class Dstack(Operator):
    """
    Diagonal stack of operators
    y1 = | A  0 |  x1
    y2   | 0  B |  x2
    """
    
    def __init__(self, *args):
        """Constructor for the stacked operator"""
        
        self.ops = []
        for _, arg in enumerate(args):
            if arg is None:
                continue
            elif isinstance(arg, Operator):
                self.ops.append(arg)
            elif isinstance(arg, list):
                for op in arg:
                    if op is None:
                        continue
                    elif isinstance(op, Operator):
                        self.ops.append(op)
            else:
                raise TypeError('Argument must be either Operator or list of Operators')
        
        # build domain and range
        self.n = len(self.ops)
        op_range = []
        op_domain = []
        for idx in range(self.n):
            op_domain += [self.ops[idx].domain]
            op_range += [self.ops[idx].range]
        
        super(Dstack, self).__init__(domain=superVector(op_domain), range=superVector(op_range))
    
    def __str__(self):
        return " DStack "
    
    def forward(self, add, model, data):
        """Forward operator"""
        self.checkDomainRange(model, data)
        for idx in range(self.n):
            self.ops[idx].forward(add, model.vecs[idx], data.vecs[idx])
    
    def adjoint(self, add, model, data):
        """Adjoint operator"""
        self.checkDomainRange(model, data)
        for idx in range(self.n):
            self.ops[idx].adjoint(add, model.vecs[idx], data.vecs[idx])


class _sumOperator(Operator):
    """
    Sum of two operators
        C = A + B
        C.H = A.H + B.H
    """
    
    def __init__(self, A, B):
        """Sum operator constructor"""
        if not isinstance(A, Operator) or not isinstance(B, Operator):
            raise TypeError('Both operands have to be a Operator')
        if not A.range.checkSame(B.range) or not A.domain.checkSame(B.domain):
            raise ValueError('Cannot add operators: shape mismatch')
        
        super(_sumOperator, self).__init__(A.domain, A.range)
        self.args = (A, B)
    
    def __str__(self):
        return self.args[0].__str__()[:3] + "+" + self.args[1].__str__()[:4]
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        self.args[0].forward(add, model, data)
        self.args[1].forward(True, model, data)
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        self.args[0].adjoint(add, model, data)
        self.args[1].adjoint(True, model, data)
 

class _prodOperator(Operator):
    """
    Multiplication of two operators
    C = A * B
    C.H = B.H * A.H
    """
    
    def __init__(self, A, B):
        if not isinstance(A, Operator) or not isinstance(B, Operator):
            raise TypeError('Both operands have to be a Operator')
        if not A.domain.checkSame(B.range):
            raise ValueError('Cannot multiply operators: shape mismatch')
        super(_prodOperator, self).__init__(B.domain, A.range)
        self.args = (A, B)
        self.temp = B.getRange().clone()
    
    def __str__(self):
        return self.args[0].__str__()[:3] + "*" + self.args[1].__str__()[:4]
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        self.args[1].forward(False, model, self.temp)
        self.args[0].forward(add, self.temp, data)
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        self.args[0].adjoint(False, self.temp, data)
        self.args[1].adjoint(add, model, self.temp)


class _scaledOperator(Operator):
    """
    Scalar matrix multiplication
    """
    
    def __init__(self, A, const):
        if not isinstance(A, Operator):
            raise TypeError('Operator expected as A')
        if not type(const) in [int, float]:
            raise ValueError('scalar expected as const')
        super(_scaledOperator, self).__init__(A.domain, A.range)
        self.const = const
        self.op = A
    
    def __str__(self):
        op_name = self.op.__str__().replace(" ", "")
        l = len(op_name)
        if l <= 6:
            name = "sc" + op_name + "" * (6 - l)
        else:
            name = "sc" + op_name[:6]
        return name
    
    def forward(self, add, model, data):
        self.op.forward(add, model.clone().scale(self.const), data)
    
    def adjoint(self, add, model, data):
        self.op.adjoint(add, model, data.clone().scale(np.conj(self.const)))


def Chain(A, B):
    """
         Chain of two operators
                d = B A m
    """
    return _prodOperator(B, A)


# for backward compatibility
Transpose = Operator.H
sumOperator = _sumOperator
stackOperator = Vstack



import numpy as np
import os
import dask.distributed as daskD

from occamypy.utils.os import BUF_SIZE
from occamypy.utils import sep
from occamypy import Vector, VectorNumpy
from .utils import DaskClient

# Verify if SepVector modules are presents
try:
    import SepVector
    
    
    def call_constr_hyper(axes_in):
        """Function to remotely construct an SepVector using the axis object"""
        return SepVector.getSepVector(axes=axes_in)
except ImportError:
    SepVector = None


def _copy_from_NdArray(vecObj, NdArray):
    """Function to set vector values from numpy array"""
    vecObj.getNdArray()[:] = NdArray
    return


# Functions necessary to submit method calls using Dask client
def _call_getNdArray(vecObj):
    """Function to call getNdArray method"""
    res = vecObj.getNdArray()
    return res


def _call_shape(vecObj):
    """Function to return shape attribute"""
    res = vecObj.shape
    return res


def _call_ndim(vecObj):
    """Function to return ndim attribute"""
    res = vecObj.ndim
    return res


def _call_size(vecObj):
    """Function to return size attribute"""
    res = vecObj.size
    return res


def _call_norm(vecObj, N=2):
    """Function to call norm method"""
    res = vecObj.norm(N)
    return res


def _call_zero(vecObj):
    """Function to call zero method"""
    res = vecObj.zero()
    return res


def _call_max(vecObj):
    """Function to call max method"""
    res = vecObj.max()
    return res


def _call_min(vecObj):
    """Function to call min method"""
    res = vecObj.min()
    return res


def _call_set(vecObj, val):
    """Function to call set method"""
    res = vecObj.set(val)
    return res


def _call_scale(vecObj, sc):
    """Function to call scale method"""
    res = vecObj.scale(sc)
    return res


def _call_addbias(vecObj, bias):
    """Function to call addbias method"""
    res = vecObj.addbias(bias)
    return res


def _call_rand(vecObj):
    """Function to call rand method"""
    res = vecObj.rand()
    return res


def _call_clone(vecObj):
    """Function to call clone method"""
    res = vecObj.clone()
    return res


def _call_cloneSpace(vecObj):
    """Function to call cloneSpace method"""
    res = vecObj.cloneSpace()
    return res


def _call_checkSame(vecObj, vec2):
    """Function to call cloneSpace method"""
    res = vecObj.checkSame(vec2)
    return res


def _call_writeVec(vecObj, filename, mode):
    """Function to call cloneSpace method"""
    res = vecObj.writeVec(filename, mode)
    return res


def _call_abs(vecObj):
    """Function to call abs method"""
    res = vecObj.abs()
    return res


def _call_sign(vecObj):
    """Function to call sign method"""
    res = vecObj.sign()
    return res


def _call_reciprocal(vecObj):
    """Function to call reciprocal method"""
    res = vecObj.reciprocal()
    return res


def _call_maximum(vecObj, vec2):
    """Function to call maximum method"""
    res = vecObj.maximum(vec2)
    return res


def _call_conj(vecObj):
    """Function to call conj method"""
    res = vecObj.conj()
    return res


def _call_real(vecObj):
    """Function to call real method"""
    res = vecObj.real()
    return res


def _call_imag(vecObj):
    """Function to call imag method"""
    res = vecObj.imag()
    return res


def _call_pow(vecObj, power):
    """Function to call pow method"""
    res = vecObj.pow(power)
    return res


def _call_copy(vecObj, vec2):
    """Function to call copy method"""
    res = vecObj.copy(vec2)
    return res


def _call_scaleAdd(vecObj, vec2, sc1, sc2):
    """Function to call scaleAdd method"""
    res = vecObj.scaleAdd(vec2, sc1, sc2)
    return res


def _call_dot(vecObj, vec2):
    """Function to call dot method"""
    res = vecObj.dot(vec2)
    return res


def _call_multiply(vecObj, vec2):
    """Function to call multiply method"""
    res = vecObj.multiply(vec2)
    return res


def _call_isDifferent(vecObj, vec2):
    """Function to call isDifferent method"""
    res = vecObj.isDifferent(vec2)
    return res


def _call_clipVector(vecObj, low, high):
    """Function to call multiply method"""
    res = vecObj.clipVector(low, high)
    return res


# Check consistency between vectors
def checkVector(vec1, vec2):
    """Function to check type and chunks of Dask-vector objects"""
    if type(vec1) is not DaskVector:
        raise TypeError("Self vector is not a DaskVector")
    if type(vec2) is not DaskVector:
        raise TypeError("Input vector is not a DaskVector")
    Nvec1 = len(vec1.vecDask)
    Nvec2 = len(vec2.vecDask)
    if Nvec1 != Nvec2:
        raise ValueError("Number of chunks is different! (self chunks %s; vec2 chunks %s)" % (Nvec1, Nvec2))
    return


class DaskVector(Vector):
    """Definition of a vector object whose computations are performed through a Dask Client"""
    
    def __init__(self, dask_client, **kwargs):
        """
        Dask Vector constructor
        dask_client = [no default] - DaskClient; client object to use when submitting tasks (see dask_util module)
        kwargs:
         - vector_template = [no default] - vector class; Vector to use to create chunks of vectors
         - chunks          = [no default] - list; List defininig the size of the multiple instances of the vector template
         or
         - vectors         = [no default] - list; List containing vectors to be spread across Dask workers
         - copy            = [True] - boolean; Whether to copy the content of the vectors or not
         - chunks          = [None] - list; List defininig how the vector list should be spread; if not specified the vectors will be evenly distributed
         or
         - dask_vectors    = [no default] - list; List containing pointers to futures to vector object (useful for clone function)
        """
        # Client to submit tasks
        super(DaskVector).__init__()
        if not isinstance(dask_client, DaskClient):
            raise TypeError("Passed client is not a Dask Client object!")
        self.dask_client = dask_client
        self.client = self.dask_client.getClient()
        # List containing futures to vectors
        self.vecDask = []
        # Getting worker IDs
        wrkIds = self.dask_client.getWorkerIds()
        N_wrk = self.dask_client.getNworkers()
        if "vector_template" in kwargs and "chunks" in kwargs:
            vec_tmplt = kwargs.get("vector_template")
            self.chunks = kwargs.get("chunks")
            # Spreading chunks across available workers
            self. chunks = [np.sum(ix) for ix in np.array_split(self.chunks, N_wrk)]
            # Checking if an SepVector was passed (by getting Hypercube)
            hyper = False
            if SepVector:
                if isinstance(vec_tmplt, SepVector.vector):
                    hyper = True
            # Broadcast vector space
            if hyper:
                vec_space = vec_tmplt.getHyper().axes  # Passing axes since Hypercube cannot be serialized
            else:
                vec_space = vec_tmplt.cloneSpace()
            vec_spaceD = self.client.scatter(vec_space, workers=wrkIds)
            daskD.wait(vec_spaceD)
            # Spreading vectors
            for iwrk, wrkId in enumerate(wrkIds):
                for ivec in range(self.chunks[iwrk]):
                    if hyper:
                        # Instantiating Sep vectors on remote machines
                        self.vecDask.append(
                            self.client.submit(call_constr_hyper, vec_spaceD, workers=[wrkId], pure=False))
                    else:
                        # Scattering vector to different workers
                        self.vecDask.append(
                            self.client.submit(_call_clone, vec_spaceD, workers=[wrkId], pure=False))
        elif "vectors" in kwargs:
            # Vector list to be spread across workers
            vec_list = kwargs.get("vectors")
            copy = kwargs.get("copy", True)
            self.chunks = kwargs.get("chunks", None)
            if self.chunks is None:
                # Spread vectors evenly
                vec_chunks = np.array_split(vec_list, N_wrk)
            else:
                # Spread according to chunk size
                if len(vec_list) != np.sum(self.chunks):
                    raise ValueError("Total number of vectors in chunks not consistent with number of vectors!")
                # Spreading chunks across available workers
                self.chunks = [np.sum(ix) for ix in np.array_split(self.chunks, N_wrk)]
                vec_chunks = np.split(vec_list, np.cumsum(self.chunks))[:-1]
            # Spreading vectors
            for iwrk, wrkId in enumerate(wrkIds):
                for vec in vec_chunks[iwrk]:
                    # Checking if an SepVector was passed
                    IsSepVec = False
                    if SepVector:
                        if isinstance(vec, SepVector.vector):
                            IsSepVec = True
                    if IsSepVec:
                        # Instantiating Sep vectors on remote machines
                        self.vecDask.append(
                            self.client.submit(call_constr_hyper, vec.getHyper().axes, workers=[wrkId], pure=False))
                        # Copying values from NdArray (Cannot scatter SepVector)
                        daskD.wait(self.vecDask[-1])
                        if copy:
                            arrD = self.client.scatter(vec.getNdArray(), workers=[wrkId])
                            daskD.wait(arrD)
                            daskD.wait(
                                self.client.submit(_copy_from_NdArray, self.vecDask[-1], arrD, pure=False))
                    else:
                        if copy:
                            self.vecDask.append(self.client.scatter(vec, workers=[wrkId]))
                        else:
                            self.vecDask.append(
                                self.client.submit(_call_clone, vec.cloneSpace(), workers=[wrkId], pure=False))
        elif "dask_vectors" in kwargs:
            dask_vectors = kwargs.get("dask_vectors")
            for dask_vec in dask_vectors:
                if not issubclass(dask_vec.type, Vector):
                    raise TypeError("One instance in dask_vectors is not a vector-derived object!")
            self.dask_client = dask_client
            self.client = self.dask_client.getClient()
            self.vecDask = dask_vectors
        else:
            raise ValueError("Wrong arguments passed to constructor! Please, read object help!")
        # Waiting vectors to be instantiated
        daskD.wait(self.vecDask)
        return
    
    # Class vector operations
    def getNdArray(self):
        """
        Function to return a list of all the arrays of the vector
        """
        futures = self.client.map(_call_getNdArray, self.vecDask, pure=False)
        arrays = self.client.gather(futures)
        return arrays
    
    @property
    def shape(self):
        futures = self.client.map(_call_shape, self.vecDask, pure=False)
        shapes = self.client.gather(futures)
        return shapes
    
    @property
    def size(self):
        """Attribute of total number of elements in the vector"""
        futures = self.client.map(_call_size, self.vecDask, pure=False)
        sizes = self.client.gather(futures)
        return np.sum(sizes)
    
    @property
    def ndim(self):
        """Attribute of number of dimensions"""
        futures = self.client.map(_call_ndim, self.vecDask, pure=False)
        ndims = self.client.gather(futures)
        return ndims
    
    def norm(self, N=2):
        """Function to compute vector N-norm"""
        norms = self.client.map(_call_norm, self.vecDask, N=N, pure=False)
        norm = 0.0
        for future, result in daskD.as_completed(norms, with_results=True):
            norm += np.power(np.float64(result), N)
        return np.power(norm, 1. / N)
    
    def zero(self):
        """Function to zero out a vector"""
        daskD.wait(self.client.map(_call_zero, self.vecDask, pure=False))
        return self
    
    def max(self):
        """Function to obtain maximum value within a vector"""
        maxs = self.client.map(_call_max, self.vecDask, pure=False)
        max_val = - np.inf
        for future, result in daskD.as_completed(maxs, with_results=True):
            if result > max_val:
                max_val = result
        return max_val
    
    def min(self):
        """Function to obtain minimum value within a vector"""
        mins = self.client.map(_call_min, self.vecDask, pure=False)
        min_val = np.inf
        for future, result in daskD.as_completed(mins, with_results=True):
            if result < min_val:
                min_val = result
        return min_val
    
    def set(self, val):
        """Function to set all values in the vector"""
        daskD.wait(self.client.map(_call_set, self.vecDask, val=val, pure=False))
        return self
    
    def scale(self, sc):
        """Function to scale a vector"""
        daskD.wait(self.client.map(_call_scale, self.vecDask, sc=sc, pure=False))
        return self
    
    def addbias(self, bias):
        """Function to add bias to a vector"""
        daskD.wait(self.client.map(_call_addbias, self.vecDask, bias=bias, pure=False))
        return self
    
    def rand(self):
        """Function to randomize a vector"""
        daskD.wait(self.client.map(_call_rand, self.vecDask, pure=False))
        return self
    
    def clone(self):
        """Function to clone (deep copy) a vector from a vector or a Space"""
        vectors = self.client.map(_call_clone, self.vecDask, pure=False)
        daskD.wait(vectors)
        return DaskVector(self.dask_client, dask_vectors=vectors)
    
    def cloneSpace(self):
        """Function to clone vector space"""
        vectors = self.client.map(_call_cloneSpace, self.vecDask, pure=False)
        daskD.wait(vectors)
        return DaskVector(self.dask_client, dask_vectors=vectors)
    
    def checkSame(self, other):
        """Function to check to make sure the vectors exist in the same space"""
        checkVector(self, other)
        futures = self.client.map(_call_checkSame, self.vecDask, other.vecDask, pure=False)
        results = self.client.gather(futures)
        return all(results)
    
    def writeVec(self, filename, mode='w', multi_file=False):
        """
        Function to write vector to file:

        :param filename     : string - Filename to write the vector to
        :param mode         : string - Writing mode 'w'=overwrite file or 'a'=append to file ['w']
        :param multi_file   : boolean - If True multiple files will be written with suffix _chunk1,2,3,...;
                              otherwise, a single will be written [False]
        """
        # Check writing mode
        if not mode in 'wa':
            raise ValueError("Mode must be appending 'a' or writing 'w' ")
        # Multi-node writing mode
        Nvecs = len(self.vecDask)
        # Creating vector-chunk names
        vec_names = [os.getcwd() + "/" + "".join(filename.split('.')[:-1]) + "_chunk%s.H" % (ii + 1)
                     for ii in range(Nvecs)]
        futures = self.client.map(_call_writeVec, self.vecDask, vec_names, [mode] * Nvecs, pure=False)
        daskD.wait(futures)
        # Single-file writing mode (concatenating all binary files)
        if not multi_file:
            # Getting binary-file locations
            bin_files = [sep.get_binary(vec_name) for vec_name in vec_names]
            # Getting all-axis information
            ax_info = [sep.get_axes(vec_name)[:sep.get_num_axes(vec_name)] for vec_name in vec_names]
            binfile = sep.datapath + filename.split('/')[-1] + '@'
            # Checks for writing header file
            len_ax = [len(ax) for ax in ax_info]
            max_len_idx = np.argmax(len_ax)
            cat_axis_multi = len_ax[max_len_idx]  # Axis on with files are concatenated
            # Getting largest-vector-axis information
            main_axes = ax_info[max_len_idx]
            N_elements_multi = 0  # Number of elements on the concatenation axis of multifiles
            # Getting number of elements if appending mode is requested
            last_axis = [[1, 1.0, 1.0, "Undefined"]]
            if os.path.isfile(filename) and 'a' in mode:
                file_axes = sep.get_axes(filename)
                last_axis[0][0] += file_axes[cat_axis_multi][0]
            # Checking compatibility of vectors
            for axes2check in ax_info:
                # First checking for len of given axis
                Naxes = len(axes2check)
                if Naxes < cat_axis_multi - 1:
                    print("WARNING! Cannot write single file with given vector chunks: "
                          "number of axes not compatible. Wrote chunks!")
                    return
                for idx, ax in enumerate(axes2check):
                    if ax[0] != main_axes[idx][0] and idx != cat_axis_multi - 1:
                        print("WARNING! Cannot write single file with given vector chunks: "
                              "elements on axis number %s not compatible. Wrote chunks!" % (idx + 1))
                        return
                if Naxes == cat_axis_multi:
                    N_elements_multi += axes2check[cat_axis_multi - 1][
                        0]  # Adding number of elements on the given concatenation axis
                else:
                    N_elements_multi += 1  # Only one element present
            # Changing number of elements on last axis
            main_axes[-1][0] = N_elements_multi
            # Adding last appending axes if file existed
            main_axes += last_axis
            # Writing header file
            with open(filename, mode) as fid:
                for ii, ax in enumerate(main_axes):
                    ax_id = ii + 1
                    fid.write("n%s=%s o%s=%s d%s=%s label%s='%s'\n" % (
                        ax_id, ax[0], ax_id, ax[1], ax_id, ax[2], ax_id, ax[3]))
                fid.write("in='%s'\n" % binfile)
                fid.write("esize=4\n")
                fid.write("data_format=\"native_float\"\n")
            # Writing binary file ("reading each binary file by chuncks of BUF_SIZE")
            with open(binfile, mode + 'b') as fid:
                for binfile_ii in bin_files:
                    with open(binfile_ii, 'rb') as fid_toread:
                        while True:
                            data = fid_toread.read(BUF_SIZE)
                            if not data: break
                            fid.write(data)
            # Removing header binary files associated to chunks
            for idx, vec_name in enumerate(vec_names):
                os.remove(vec_name)
                os.remove(bin_files[idx])
        return
    
    def abs(self):
        """Return a vector containing the absolute values"""
        daskD.wait(self.client.map(_call_abs, self.vecDask, pure=False))
        return self
    
    def sign(self):
        """Return a vector containing the signs"""
        daskD.wait(self.client.map(_call_sign, self.vecDask, pure=False))
        return self
    
    def reciprocal(self):
        """Return a vector containing the reciprocals of self"""
        daskD.wait(self.client.map(_call_reciprocal, self.vecDask, pure=False))
        return self
    
    def conj(self):
        """Compute conjugate transpose of the vector"""
        daskD.wait(self.client.map(_call_conj, self.vecDask, pure=False))
        return self
    
    def real(self):
        """Return the real part of the vector"""
        daskD.wait(self.client.map(_call_real, self.vecDask, pure=False))
        return self
    
    def imag(self):
        """Return the imaginary part of the vector"""
        daskD.wait(self.client.map(_call_imag, self.vecDask, pure=False))
        return self
    
    def pow(self, power):
        """Compute element-wise power of the vector"""
        daskD.wait(self.client.map(_call_pow, self.vecDask, power=power, pure=False))
        return self
    
    # Methods combinaning different vectors
    
    def maximum(self, vec2):
        """Return a new vector of element-wise maximum of self and vec2"""
        checkVector(self, vec2)
        daskD.wait(self.client.map(_call_maximum, self.vecDask, vec2.vecDask, pure=False))
        return self
    
    def copy(self, other):
        """Function to copy vector"""
        checkVector(self, other)
        daskD.wait(self.client.map(_call_copy, self.vecDask, other.vecDask, pure=False))
        return self
    
    def scaleAdd(self, other, sc1=1.0, sc2=1.0):
        """Function to scale two vectors and add them to the first one"""
        checkVector(self, other)
        sc1 = [sc1] * len(self.vecDask)
        sc2 = [sc2] * len(self.vecDask)
        futures = self.client.map(_call_scaleAdd, self.vecDask, other.vecDask, sc1, sc2,
                                  pure=False)
        daskD.wait(futures)
        return self
    
    def dot(self, other):
        """Function to compute dot product between two vectors"""
        checkVector(self, other)
        dots = self.client.map(_call_dot, self.vecDask, other.vecDask, pure=False)
        # Adding all the results together
        dot = 0.0
        for future, result in daskD.as_completed(dots, with_results=True):
            dot += np.float64(result)
        return dot
    
    def multiply(self, other):
        """Function to multiply element-wise two vectors"""
        checkVector(self, other)
        futures = self.client.map(_call_multiply, self.vecDask, other.vecDask, pure=False)
        daskD.wait(futures)
        return self
    
    def isDifferent(self, vec2):
        """Function to check if two vectors are identical"""
        checkVector(self, vec2)
        futures = self.client.map(_call_isDifferent, self.vecDask, vec2.vecDask,
                                  pure=False)
        results = self.client.gather(futures)
        return any(results)
    
    def clipVector(self, low, high):
        """Function to bound vector values based on input vectors min and max"""
        checkVector(self, low)  # Checking low-bound vector
        checkVector(self, high)  # Checking high-bound vector
        futures = self.client.map(_call_clipVector, self.vecDask, low.vecDask,
                                  high.vecDask, pure=False)
        daskD.wait(futures)
        return self


# DASK I/O TO READ LARGE-SCALE VECTORS DIRECTLY WITHIN EACH WORKER
def _get_binaries(**kwargs):
    """
    Function to obtain associated binary files to each file name
    :param filenames: list; List/Array containing file names to read
    :return:
    binfiles: list; List containing binary files associated to each file
    Nbytes: list; List containing the number of bytes within binary files
    """
    binfiles = list()
    Nbytes = list()
    filenames = kwargs.get("filenames")
    for filename in filenames:
        _, ext = os.path.splitext(filename)  # Getting file extension
        if ext == ".H":  # SEPlib file
            binfiles.append(sep.get_binary(filename))
            Nbytes.append(os.path.getsize(binfiles[-1]))
        elif ext == ".h5":
            raise NotImplementedError("ERROR! h5 files not supported yet.")
        else:
            raise ValueError("ERROR! Unknown format for file %s" % filename)
    return binfiles, Nbytes


def _set_binfiles(binfiles, Nbytes, **kwargs):
    """
    Function to associate binary file/s for each vector
    :param shapes: list/array; List/Array containing the shape of each chunk to be read
    :param binfiles: list; List containing the number of bytes within binary files
    :param Nbytes: list; List containing the number of bytes within binary files
    :param format: string; Kind of binary format to read ['>f']
    :return:
    bin_chunks: list of lists; binary files associated to each vector
    counts: list of lists; number of bytes to read per binary file
    offsets: list of lists; offset to apply when reading given binary file
    """
    shapes = kwargs.get("shapes")
    fmt = kwargs.get("format", ">f")  # Default floating point number
    esize = np.dtype(fmt).itemsize  # Byte size per vector element
    # Setting bin_chunks, counts, and offsets
    bin_chunks = list()
    counts = list()
    offsets = list()
    # Checking total bytes and number of elements
    totalBytes = np.int(np.sum(Nbytes))
    rqstBytes = 0
    for shp in shapes:
        rqstBytes += np.prod(shp) * esize
    if rqstBytes > totalBytes:
        raise ValueError(
            "ERROR! Total number of bytes needed to be read (%d) greater than bytes in provided file list (%d)"
            % (rqstBytes, totalBytes))
    bytesRd = 0  # Number of bytes read so far from a given file
    for shp in shapes:
        Nelmnt = np.prod(shp)  # Total number of elements within vector
        NelmntBt = Nelmnt * esize  # Number of bytes necessary for given vector
        tmp_bin_files = list()
        tmp_counts = list()
        tmp_offsets = list()
        for _ in range(len(binfiles)):
            tmp_bin_files.append(binfiles[0])
            if NelmntBt >= Nbytes[0]:
                # Entire binary file must be read
                tmp_counts.append(np.int(Nbytes[0] / esize))
                tmp_offsets.append(bytesRd)
                # Updating variables
                bytesRd = 0
                NelmntBt -= Nbytes[0]  # Subtracting bytes already read
                # Removing the current binary file (completely read)
                binfiles.pop(0)
                Nbytes.pop(0)
            else:
                # Only part of the file needs to be read
                tmp_counts.append(np.int(NelmntBt / esize))
                tmp_offsets.append(bytesRd)
                bytesRd += NelmntBt  # Number of bytes read
                Nbytes[0] -= NelmntBt
                NelmntBt = 0  # All necessary bytes are set
            if NelmntBt == 0:
                break  # Read all the elements of a vector
        # For a given vector, placing information on bin files, counts, offset
        bin_chunks.append(tmp_bin_files)
        counts.append(tmp_counts)
        offsets.append(tmp_offsets)
    return bin_chunks, counts, offsets


def _read_vector_dask(shape, binaries, counts, offsets, **kwargs):
    """
    Function to read a vector using a Dask worker
    :param shape : - list/array; Shape of the vector to be instantiated
    :param binaries : - list; Binary files to read to instantiate a vector
    :param counts : - list; Number of elements to read per binary file
    :param offsets : -list; Bytes to be skipped when reading given file
    :return:
    vector - vector class; vector instance creates using
    """
    vector = None
    vtype = kwargs.get("vtype")
    axes = kwargs.get("axes", None)
    fmt = kwargs.get("format", ">f")  # Default floating point number
    # Reading binary data into data array
    data = np.array([])  # Initialize an empty array
    for ii, filename in enumerate(binaries):
        _, ext = os.path.splitext(filename)  # Getting file extension
        if ext == ".H@":
            fid = open(filename, 'r+b')
            # Default formatting big-ending floating point number
            data = np.append(data, np.fromfile(fid, count=counts[ii], offset=offsets[ii], dtype=fmt))
            fid.close()
        else:
            raise ValueError("ERROR! Unknown extension for binary file: %s" % filename)
    # Reshaping array and forcing memory continuity
    data = np.ascontiguousarray(np.reshape(data, shape))
    if vtype == "VectorNumpy":
        vector = VectorNumpy(data)
    elif vtype == "SepVector":
        if SepVector:
            if axes:
                # Instantiate using axis
                vector = SepVector.getSepVector(axes=axes)
            else:
                # Instantiate using shape
                shape.reverse()  # Reversing order of axis
                vector = SepVector.getSepVector(ns=shape)
            vector.getNdArray()[:] = data  # Copying data into SepVector instance
            del data  # Removing data
        else:
            raise ImportError("ERROR! SepVector module not found!")
    else:
        raise ValueError("ERROR! Unknown vtype (%s)" % vtype)
    return vector


def readDaskVector(dask_client, **kwargs):
    """
    Function to read files in parallel and store within Dask vector
    :param dask_client : - DaskClient; client object to use when submitting tasks (see dask_util module)
    :param filenames : - list; List/Array containing file names to read
    :param shapes : - list/array; List/Array containing the shape of each chunk to be read
    :param chunks : - list/array; List/Array defining how each vector should be spread. Note that len(chunks) must be
                                  equal to dask_client.getNworkers(), len(shape) must be equal np.sum(chunks)
    :param vtype : - string; Type of vectors to be instantiated. Supported (VectorNumpy, SepVector)
    :return:
    daskVec - Dask Vector object
    """
    # Getting dask client components
    Nwrks = dask_client.getNworkers()
    client = dask_client.getClient()
    wrkIds = dask_client.getWorkerIds()
    # Args
    shapes = kwargs.get("shapes")
    chunks = kwargs.get("chunks")
    vtype = kwargs.get("vtype")
    axes = kwargs.get("axes", None)  # Necessary for SepVector
    # Q/C steps
    if len(chunks) != Nwrks:
        raise ValueError(
            "ERROR! Number of workers in client (%d) inconsistent with chunks length (%d)" % (Nwrks, len(chunks)))
    if np.sum(chunks) != len(shapes):
        raise ValueError(
            "ERROR! Total number of chunks (%s) inconsistent with shapes length (%d)" % (np.sum(chunks), len(shapes)))
    if vtype not in "VectorIC SepVector":
        raise ValueError("ERROR! Provided vtype (%s) not currently supported" % vtype)
    if axes:
        if len(axes) != len(shapes):
            raise ValueError(
                "ERROR! Length of axes (%s) not consistent with length of shapes (%s)" % (len(axes), len(shapes)))
    # Pre-processing: for each chunk associate necessary files, bytes to read (count), offset (if file goes onto two
    # or more chunks).
    # Get binary files (Necessary to use header-based formats)
    binfiles, Nbytes = _get_binaries(**kwargs)
    # Associate binary files with each vector using shapes and
    # set count and offset for each of them
    bin_shps, counts, offsets = _set_binfiles(binfiles, Nbytes, **kwargs)
    # Split lists/arrays into chunks for parallel processing
    shps2read = list()
    bin2read = list()
    count2read = list()
    offset2read = list()
    # Preprocessing for vector-specific arguments
    # must be done in the next for loop
    axes2read = list()  # List of axes
    for chunk in chunks:
        shps2read.append(shapes[:chunk])
        bin2read.append(bin_shps[:chunk])
        count2read.append(counts[:chunk])
        offset2read.append(offsets[:chunk])
        del shapes[:chunk], bin_shps[:chunk], counts[:chunk], offsets[:chunk]
        # Checking if axis list was provided
        if axes and vtype == "SepVector":
            axes2read.append(axes[:chunk])
            del axes[:chunk]
    # Loop over workers/chunks
    # Read binary files and place within vector objects
    fut_vec = []
    for iWrk, wrkId in enumerate(wrkIds):
        for ivec in range(len(shps2read[iWrk])):
            if len(axes2read) > 0 and vtype == "SepVector":
                kwargs.update({"axes": axes2read[iWrk][ivec]})
            fut_vec.append(client.submit(_read_vector_dask, shps2read[iWrk][ivec], bin2read[iWrk][ivec],
                                         count2read[iWrk][ivec], offset2read[iWrk][ivec], **kwargs,
                                         workers=[wrkId], pure=False))
    # Waiting for all vector chunks to be instantiated
    daskD.wait(fut_vec)
    # Checking for errors
    for idx, vec in enumerate(fut_vec):
        if vec.status == 'error':
            print("Error for when instantiating vector %s" % idx)
            print(vec.result())
    daskVec = DaskVector(dask_client, dask_vectors=fut_vec)
    return daskVec



import numpy as np
from .base import Operator


class Zero(Operator):
    """Zero matrix operator; useful for Jacobian matrices that are zeros"""
    
    def __init__(self, domain, range):
        super(Zero, self).__init__(domain, range)
    
    def __str__(self):
        return "  Zero  "
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            model.zero()


class Identity(Operator):
    """Identity operator"""
    
    def __init__(self, domain):
        super(Identity, self).__init__(domain, domain)
    
    def __str__(self):
        return "Identity"
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        if add:
            data.scaleAdd(model)
        else:
            data.copy(model)
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        if add:
            model.scaleAdd(data)
        else:
            model.copy(data)


class Scaling(Operator):
    """scalar multiplication operator"""
    
    def __init__(self, domain, scalar):
        super(Scaling, self).__init__(domain, domain)
        if not np.isscalar(scalar):
            raise ValueError('scalar has to be (indeed) a scalar variable')
        self.scalar = scalar
    
    def __str__(self):
        return "Scaling "
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        data.scaleAdd(model, 1. if add else 0., self.scalar)
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        model.scaleAdd(data, 1. if add else 0., self.scalar)


class Diagonal(Operator):
    """Diagonal operator for performing element-wise multiplication"""
    
    def __init__(self, diag):
        # if not isinstance(diag, vector):
        #     raise TypeError('diag has to be a vector')
        super(Diagonal, self).__init__(diag, diag)
        self.diag = diag
    
    def __str__(self):
        return "Diagonal"
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        data.scaleAdd(model, 1. if add else 0.)
        data.multiply(self.diag)
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        model.scaleAdd(data, 1. if add else 0.)
        model.multiply(self.diag)



import numpy as np
from math import floor, ceil, log2
from occamypy import Operator, VectorBaseIC


def _fft_direct(x):
    return np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(x)) / np.sqrt(np.prod(x.shape)))


def _fft_inverse(x):
    return np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(x))) * np.sqrt(np.prod(x.shape))


def _fdct_wrapping_window(x):
    """
    Create the two halves of a C^inf compactly supported window
    
    :param x : :obj:`np.ndarray´
        vector of abscissae, the relevant ones from 0 to 1
    :return wl, wr: vectors of samples of the left and right halves of the window
    """
    wl = np.zeros_like(x)
    wr = np.zeros_like(x)
    
    wr[(x > 0) & (x < 1)] = np.exp(1 - 1. / (1 - np.exp(1 - 1. / x[(x > 0) & (x < 1)])))
    wr[x <= 0] = 1
    wl[(x > 0) & (x < 1)] = np.exp(1 - 1. / (1 - np.exp(1 - 1. / (1 - x[(x > 0) & (x < 1)]))))
    wl[x >= 1] = 1
    
    norm_vector = np.sqrt(wl ** 2 + wr ** 2)
    
    return wl / norm_vector, wr / norm_vector


class Curvelet2D(Operator):
    """
    2D Fast Discrete Curvelet Transform

    :param model : :obj:`Vector`
        2D domain vector
    :param finest_curvelet : :obl:`bool`
        Use curvelets at the finest level instead of wavelets (False by default)
    :param num_scales : :obj:`int`, optional
        Number of scales including the coarsest wavelet level (ceil(log2(min(M,N))-3 by default)
    :param num_angles_coarse : :obj:`int`, optional
        Number of angles at the second coarsest wavelet level; must be >= 8 and multiple of 4 (16 by default)
    :param is_real : :obj:`bool`, optional
        Use the real-valued transform (False by default)

    Note
    ----
    Porting by Francesco Picetti based on the code of Emmanuel Candes, 2003--2004
    """
    
    def __init__(self, model, num_scales=None, num_angles_coarse=16, finest_curvelet=False, is_real=False):
        
        assert model.ndim == 2, 'model has to be a 2D Vector'
        
        self.N1, self.N2 = model.shape
        
        self.num_scales = num_scales if num_scales is not None else ceil(log2(min(N1, N2)) - 3)
        
        assert num_angles_coarse >= 8, 'num_angles_coarse has to be >= 8'
        assert num_angles_coarse % 4 == 0, 'num_angles_coarse has to be multiple of 4'
        self.num_angles_coarse = num_angles_coarse
        
        self.finest_curvelet = finest_curvelet
        self.is_real = is_real
        
        # data structure
        self.num_angles = np.insert(
            self.num_angles_coarse * 2 ** (np.ceil((self.num_scales - np.arange(self.num_scales, 1, -1)) / 2)), 0,
            1).astype(int)
        
        if not self.finest_curvelet:
            self.num_angles[self.num_scales] = 1
        
        self.C = [None] * self.num_scales
        for s in range(self.num_scales):
            self.C[s] = [None] * self.num_angles[s]
        
        self.scales = np.arange(self.num_scales, 1, -1) if self.finest_curvelet else np.arange(self.num_scales - 1, 1,
                                                                                               -1)
        
        self.num_quad = 2 if self.is_real else 4
        self.num_angles_per_quad = self.num_angles / 4
        
        super(Curvelet2D, self).__init__(domain=VectorBaseIC(np.zeros(model.shape, dtype=np.complex)),
                                         range=VectorBaseIC(np.zeros(shape=dims_fft, dtype=np.complex)))
    
    def __str__(self):
        return 'Curvelet'
    
    def _compute_filters(self, m1, m2):
        window_length_1 = floor(2 * m1) - floor(m1) - 1
        window_length_2 = floor(2 * m2) - floor(m2) - 1
        coord_1 = np.linspace(0, 1, num=window_length_1, endpoint=True)
        coord_2 = np.linspace(0, 1, num=window_length_2, endpoint=True)
        wl_1, wr_1 = _fdct_wrapping_window(coord_1)
        wl_2, wr_2 = _fdct_wrapping_window(coord_2)
        lowpass_1 = np.concatenate((wl_1, np.ones(2 * floor(m1) + 1), wr_1))
        lowpass_2 = np.concatenate((wl_2, np.ones(2 * floor(m2) + 1), wr_2))
        
        lowpass = lowpass_1.conj().T @ lowpass_2
        hipass = np.sqrt(1 - lowpass ** 2)
        return lowpass, hipass
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        modelNd = model.getNdArray()
        dataNd = data.getNdArray()
        
        # Loop: pyramidal scale decomposition
        M1 = self.N1 / 3
        M2 = self.N2 / 3
        
        if self.finest_curvelet:
            pass
        else:
            # last scale
            X = _fft_direct(modelNd)
            
            M1, M2 = M1 / 2, M2 / 2
            l1 = np.arange(-floor(2 * M1), floor(2 * M1)) + ceil((self.N1 + 1) / 2)
            l2 = np.arange(-floor(2 * M2), floor(2 * M2)) + ceil((self.N2 + 1) / 2)
            lpf, hpf = self._compute_filters(M1 / 2, M2 / 2)
            
            Xlow = X[l1, l2] * lpf
            Xhi = X.copy()
            Xhi[l1, l2] *= hpf
            
            C = self.C.copy()
            
            C[self.num_scales - 1][0] = _fft_inverse(Xhi)
            if self.is_real:
                C[self.num_scales - 1][0] = C[self.num_scales - 1][0].real
        
        for j in self.scales:
            M1 /= 2
            M2 /= 2
            Xhi = Xlow.copy()
            
            lpf, hpf = self._compute_filters(M1, M2)
            l1 = np.arange(-floor(2 * M1), floor(2 * M1)) + floor(4 * M1) + 1
            l2 = np.arange(-floor(2 * M2), floor(2 * M2)) + floor(4 * M2) + 1
            
            Xlow = Xlow[l1, l2]
            Xhi[l1, l2] = Xlow * hpf
            Xlow *= lpf
            
            # loop: angular decomposition
            l = 0
            for quadrant in range(self.num_quad):
                M_horiz = M2 * (quadrant % 2 == 1) + M1 * (quadrant % 2 == 0)
                M_vert = M1 * (quadrant % 2 == 1) + M2 * (quadrant % 2 == 0)
                
                wedge_ticks_left = np.round(
                    np.linspace(0, 0.5, num=2 * self.num_angles_per_quad[j]) * 2 * floor(4 * M_horiz) + 1)
                wedge_ticks_right = 2 * floor(4 * M_horiz) + 2 - wedge_ticks_left
                wedge_ticks = np.concatenate((wedge_ticks_left, np.flip(wedge_ticks_right)))
                if not self.num_angles_per_quad[j] % 2:
                    wedge_ticks = wedge_ticks[:-1]
                wedge_endpoints = wedge_ticks[1:2:-1]
                wedge_midpoints = (wedge_endpoints[:-1] + wedge_endpoints[1:]) / 2
                
                # left corner wedge
                l += 1
                first_wedge_endpoint_vert = round(2 * floor(4 * M_vert) / (2 * self.num_angles_per_quad) + 1)
                length_corner_wedge = floor(4 * M_vert) - floor(M_vert) + ceil(first_wedge_endpoint_vert / 4)
                Y_corner = np.arange(length_corner_wedge)
                XX, YY = np.meshgrid(np.arange(2 * floor(4 * M_horiz) + 1), Y_corner)
                
                width_wedge = wedge_endpoints[1] + wedge_endpoints[0] - 1
                slope_wedge = (floor(4 * M_horiz) + 1 - wedge_endpoints[0]) / floor(4 * M_vert)
                left_line = round(2 - wedge_endpoints[0] + slope_wedge * Y_corner)
                
                [wrapped_data, wrapped_XX, wrapped_YY] = deal(zeros(length_corner_wedge, width_wedge));
        
        return
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        modelNd = model.getNdArray()
        dataNd = data.getNdArray()
        # here we need to separate the computation and use np.take for handling nfft > model.shape
        temp = np.fft.ifftn(dataNd, s=self.nfft, axes=self.axes, norm='ortho')
        for a in self.axes:
            temp = np.take(temp, range(self.domain.shape[a]), axis=a)
        modelNd += temp
        return



# Module containing useful function to interact with the system
import subprocess
import sys
import os
import hashlib
import random
import string
from importlib.util import find_spec
from .logger import Logger


CUPY_ENABLED = find_spec("cupy") is not None

debug = False  # Debug flag for printing screen output of RunShellCmd as it runs commands
debug_log = None  # File where debug outputs are written if requested (it must be a logger object)
BUF_SIZE = 8388608  # read binary files in 64Mb chunks!
DEVNULL = open(os.devnull, 'wb')


def mkdir(directory):
    try:
        if not os.path.exists(directory):
            os.makedirs(directory)
    except OSError:
        print('Error: Creating directory. ' + directory)


def rand_name(N: int = 6) -> str:
    """function returning random sequence of N letters and numbers"""
    return ''.join(random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits) for _ in range(N))


def hashfile(binfile):
    """Function hashing a binary file. It uses a BUF_SIZE to partially store file in memory
    and do not completely load the file into the RAM"""
    md5 = hashlib.md5()
    with open(binfile, 'rb') as fid:
        while True:
            data = fid.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


def RunShellCmd(cmd, print_cmd=False, print_output=False, synch=True, check_code=True, get_stat=True, get_output=True):
    """Function to run a Shell command through python, return code and """
    # Overwrites any previous definition (when used within other programs)
    global debug, debug_log
    # Running command synchronously or asynchronously?
    if synch:
        if debug:
            print_cmd = True
            print_output = True
            # Printing command to be run if requested
        info = "RunShellCmd running: \'%s\'" % cmd
        if isinstance(debug_log, Logger):
            debug_log.addToLog(info)
        if print_cmd:
            print(info)
        # Starting the process (Using PIPE to streaming output)
        proc = subprocess.Popen([cmd], stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT, shell=True, universal_newlines=True)
        # Creating Stdout to save command output
        stdout = []
        # Streaming the stdout to screen if requested
        while True:
            line = proc.stdout.readline()
            if line == '' and proc.poll() is not None:
                break
            else:
                stdout.append(line)
                line = line.rstrip()
                if line != '':
                    # Print to debug file?
                    if isinstance(debug_log, Logger):
                        debug_log.addToLog(line)
                    # Print to screen?
                    if print_output:
                        print(line)
                    sys.stdout.flush()
                proc.stdout.flush()
        stdout = ''.join(stdout)
    else:
        # Running process asynchronously (Avoiding PIPE and removing standard output)
        global DEVNULL
        proc = subprocess.Popen([cmd], stdout=DEVNULL, shell=True, universal_newlines=True)
        return proc, "Running command asynchronously, returning process"
    # Command has finished, checking error code and returning requested variables
    err_code = proc.poll()
    return_var = []
    # Returning error code or status
    if get_stat:
        return_var.append(err_code)
    # Returning output
    if get_output:
        return_var.append(stdout)
    # Checking error code
    if check_code and err_code != 0:
        # Writing error code to debug file if any
        info = "ERROR! Command failed: %s; Error code: %s" % (cmd, err_code)
        if isinstance(debug_log, Logger):
            debug_log.addToLog(info)
        raise SystemError("ERROR! Command failed: %s; Error code: %s; Output: %s" % (cmd, err_code, stdout))
    # Returning
    return return_var



from typing import Union, List, Tuple
from itertools import accumulate, product
import numpy as np
import torch
from occamypy import superVector, Operator, Dstack
from ..vector import VectorTorch
from ..back_utils import set_backends

set_backends()


def _gaussian_kernel1d(sigma: float, order: int = 0, truncate: float = 4.) -> torch.Tensor:
    """
    Computes a 1-D Gaussian convolution kernel.
    """
    radius = int(truncate * sigma + 0.5)
    if order < 0:
        raise ValueError('order must be non-negative')
    exponent_range = torch.arange(order + 1)
    sigma2 = sigma * sigma
    x = torch.arange(-radius, radius+1)
    phi_x = torch.exp(-0.5 / sigma2 * x ** 2)
    phi_x = phi_x / phi_x.sum()

    if order == 0:
        return phi_x
    else:
        # f(x) = q(x) * phi(x) = q(x) * exp(p(x))
        # f'(x) = (q'(x) + q(x) * p'(x)) * phi(x)
        # p'(x) = -1 / sigma ** 2
        # Implement q'(x) + q(x) * p'(x) as a matrix operator and apply to the
        # coefficients of q(x)
        q = torch.zeros(order + 1)
        q[0] = 1
        D = torch.diag(exponent_range[1:], 1)  # D @ q(x) = q'(x)
        P = torch.diag(torch.ones(order)/-sigma2, -1)  # P @ q(x) = q(x) * p'(x)
        Q_deriv = D + P
        for _ in range(order):
            q = Q_deriv.dot(q)
        q = (x[:, None] ** exponent_range).dot(q)
        return q * phi_x
    

class ConvND(Operator):
    """
    ND convolution square operator in the domain space

    :param model  : [no default] - vector class; domain vector
    :param kernel : [no default] - vector class; kernel vector
    :return       : Convolution Operator
    """
    
    def __init__(self, model: VectorTorch,  kernel: Union[VectorTorch, torch.Tensor]):
        
        if isinstance(kernel, VectorTorch):
            self.kernel = kernel.getNdArray().clone()
        elif isinstance(kernel, torch.Tensor):
            self.kernel = kernel.clone()
        
        if model.ndim != self.kernel.ndim:
            raise ValueError("Domain and kernel number of dimensions mismatch")
        
        if model.device != self.kernel.device:
            raise ValueError("Domain and kernel has to live in the same device")
        
        self.kernel_size = tuple(self.kernel.shape)
        self.pad_size = tuple([k // 2 for k in self.kernel_size])

        if model.ndim == 1:
            conv = torch.nn.functional.conv_transpose1d
            corr = torch.nn.functional.conv1d
        elif model.ndim == 2:
            conv = torch.nn.functional.conv_transpose2d
            corr = torch.nn.functional.conv2d
        elif model.ndim == 3:
            conv = torch.nn.functional.conv_transpose3d
            corr = torch.nn.functional.conv3d
        else:
            raise ValueError
        
        # torch.nn functions require batch and channel dimensions
        self.conv = lambda x: conv(x.unsqueeze(0).unsqueeze(0),
                                   self.kernel.unsqueeze(0).unsqueeze(0),
                                   padding=self.pad_size).flatten(end_dim=2)

        self.corr = lambda x: corr(x.unsqueeze(0).unsqueeze(0),
                                   self.kernel.unsqueeze(0).unsqueeze(0),
                                   padding=self.pad_size).flatten(end_dim=2)
        
        super(ConvND, self).__init__(model, model)
    
    def __str__(self):
        return " ConvND "
    
    def forward(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        data[:] += self.conv(model.getNdArray())
        return
    
    def adjoint(self, add, model, data):
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        model[:] += self.corr(data.getNdArray())
        return
    

class GaussianFilter(ConvND):
    def __init__(self, model, sigma):
        """
        Gaussian smoothing operator.
        :param model : vector class;
            domain vector
        :param sigma : scalar or sequence of scalars;
            standard deviation along the model directions
        """
        self.sigma = [sigma] if isinstance(sigma, (float, int)) else sigma
        assert isinstance(self.sigma, (list, tuple))
        self.scaling = np.sqrt(np.prod(np.array(self.sigma) / np.pi))
        kernels = [_gaussian_kernel1d(s) for s in self.sigma]
        
        self.kernel = [*accumulate(kernels, lambda a, b: torch.outer(a.flatten(), b))][-1]
        self.kernel.reshape([k.numel() for k in kernels])
        super(GaussianFilter, self).__init__(model, self.kernel.to(model.device))
    
    def __str__(self):
        return "GausFilt"


def ZeroPad(model, pad):
    if isinstance(model, VectorTorch):
        return _ZeroPad(model, pad)
    elif isinstance(model, superVector):
        # TODO add the possibility to have different padding for each sub-vector
        return Dstack([_ZeroPad(v, pad) for v in model.vecs])
    else:
        raise ValueError("ERROR! Provided domain has to be either vector or superVector")


class _ZeroPad(Operator):
    
    def __init__(self, model: VectorTorch, pad: Union[int, Tuple[int], List[int]]):
        """ Zero Pad operator.

        To pad 2 values to each side of the first dim, and 3 values to each side of the second dim, use:
            pad=(2,2,3,3)
        :param model: VectorTorch class
        :param pad: scalar or sequence of scalars
            Number of samples to pad in each dimension.
            If a single scalar is provided, it is assigned to every dimension.
        """
        nd = model.ndim
        
        if isinstance(pad, (int, float)):
            pad = [pad, pad] * nd
        else:
            assert len(pad) == 2 * nd
            
        if (np.array(pad) < 0).any():
            raise ValueError('Padding must be positive or zero')
        
        self.pad = list(pad)
        
        self.padded_shape = tuple(np.asarray(model.shape) + [self.pad[i]+self.pad[i+1] for i in range(0, 2*nd, 2)])
        
        super(_ZeroPad, self).__init__(model, VectorTorch(self.padded_shape, device=model.device.index))

        self.inner_idx = [list(torch.arange(start=self.pad[0:-1:2][i], end=self.range.shape[i]-pad[1::2][i])) for i in range(nd)]
    
    def __str__(self):
        return "ZeroPad "
    
    def forward(self, add, model, data):
        """Zero padding"""
        self.checkDomainRange(model, data)
        if not add:
            data.zero()
        # torch counts the axes in reverse order
        data[:] += torch.nn.functional.pad(model.getNdArray(), self.pad[::-1], mode='constant')
        return
    
    def adjoint(self, add, model, data):
        """Extract non-zero subsequence"""
        self.checkDomainRange(model, data)
        if not add:
            model.zero()
        x = torch.Tensor([data[coord] for coord in product(*self.inner_idx)]).reshape(self.domain.shape).to(model.device)
        model[:] += x
        return
